import logging
import logging.handlers
import os
import sys
from pathlib import Path
from typing import Optional

def setup_logging(
    log_level: int = logging.INFO,
    log_file: Optional[str] = None,
    console: bool = True,
    max_bytes: int = 10 * 1024 * 1024,  # 10MB
    backup_count: int = 5,
) -> None:
    """
    Configure logging for the eubi-bridge package.

    Args:
        log_level: Logging level (e.g., logging.INFO, logging.DEBUG)
        log_file: Path to log file. If None, file logging is disabled
        console: Whether to log to console
        max_bytes: Maximum log file size before rotation
        backup_count: Number of backup log files to keep
    """
    # Create formatter
    formatter = logging.Formatter(
        '%(asctime)s - %(name)s - %(levelname)s - %(message)s',
        datefmt='%Y-%m-%d %H:%M:%S'
    )

    # Get the root logger
    logger = logging.getLogger('eubi_bridge')
    logger.setLevel(log_level)

    # Clear any existing handlers
    for handler in logger.handlers[:]:
        logger.removeHandler(handler)

    # Add console handler if requested
    if console:
        console_handler = logging.StreamHandler(sys.stdout)
        console_handler.setFormatter(formatter)
        logger.addHandler(console_handler)

    # Add file handler if log file is specified
    if log_file:
        # Create log directory if it doesn't exist
        log_path = Path(log_file).parent
        log_path.mkdir(parents=True, exist_ok=True)
        
        file_handler = logging.handlers.RotatingFileHandler(
            log_file,
            maxBytes=max_bytes,
            backupCount=backup_count,
            encoding='utf-8'
        )
        file_handler.setFormatter(formatter)
        logger.addHandler(file_handler)

    # Set up exception hook to log uncaught exceptions
    def handle_exception(exc_type, exc_value, exc_traceback):
        if issubclass(exc_type, KeyboardInterrupt):
            # Call default handler for KeyboardInterrupt
            sys.__excepthook__(exc_type, exc_value, exc_traceback)
            return
        
        logger.critical(
            "Uncaught exception",
            exc_info=(exc_type, exc_value, exc_traceback)
        )

    sys.excepthook = handle_exception


def get_logger(name: str) -> logging.Logger:
    """
    Get a logger with the specified name, scoped under 'eubi_bridge'.
    
    Args:
        name: Name of the logger (usually __name__)
        
    Returns:
        Configured logger instance
    """
    # Remove any existing handlers from the root logger to prevent duplicate logs
    root_logger = logging.getLogger()
    for handler in root_logger.handlers[:]:
        if not isinstance(handler, logging.NullHandler):
            root_logger.removeHandler(handler)
    
    # Return a namespaced logger
    return logging.getLogger(f'eubi_bridge.{name}')



from eubi_bridge import ebridge
from eubi_bridge import ebridge_base
from eubi_bridge import fileset_io
from eubi_bridge.base import scale, writers
from eubi_bridge.ngff import defaults, multiscales
from eubi_bridge.utils import convenience, dask_client_plugins
from eubi_bridge.utils.logging_config import setup_logging, get_logger
import logging

# Set up default logging configuration when the package is imported
setup_logging(
    log_level=logging.INFO,
    log_file=None,  # Set to a path to enable file logging
    console=True
)

# Create a logger for the package
logger = get_logger(__name__)
logger.debug("eubi_bridge package initialized")

__all__ = [
    'ebridge',
    'ebridge_base',
    'fileset_io',
    'scale',
    'writers',
    'defaults',
    'multiscales',
    'convenience',
    'dask_client_plugins',
    'setup_logging',
    'get_logger',
]



import fsspec
import fsspec.core
import fsspec.compression
import fsspec.spec
import numpy as np

from dask import delayed
import dask
import dask.array as da
from eubi_bridge.ngff.multiscales import Pyramid

readable_formats = ('.ome.tiff', '.ome.tif', '.czi', '.lif',
                    '.nd2', '.tif', '.tiff', '.lsm',
                    '.png', '.jpg', '.jpeg')


def read_tiff(input_path, **kwargs):
    from bioio_tifffile.reader import Reader as reader  # pip install bioio-tifffile --no-deps
    kwargs['chunk_dims'] = 'YX'
    img = reader(input_path, **kwargs)
    dimensions_present = img.standard_metadata.dimensions_present
    dimensions_to_read = 'TCZYX'
    if 'S' in dimensions_present and 'C' in dimensions_to_read and 'S' not in dimensions_to_read:
        dimensions_to_read = dimensions_to_read.replace('C', 'S')
    im = img.get_image_dask_data(dimensions_to_read)
    return im


@delayed
def read_single_image_asarray(input_path, **kwargs):
    """
    Reads a single image file with Dask and returns the array.

    Parameters
    ----------
    input_path : str
        Path to the image file.
    **kwargs : dict
        Additional keyword arguments, such as `verbose` and `scene`.

    Returns
    -------
    arr : dask.array.Array
        The image array.
    """
    from eubi_bridge.utils.logging_config import get_logger
    logger = get_logger(__name__)
    reader_kwargs = {}
    dimensions = 'TCZYX'
    if input_path.endswith('.zarr'):
        reader = Pyramid().from_ngff
    elif input_path.endswith(('ome.tiff', 'ome.tif')):
        from bioio_ome_tiff.reader import Reader as reader # pip install bioio-ome-tiff --no-deps
    elif input_path.endswith(('.tif', '.tiff', '.lsm')):
        reader = read_tiff
    elif input_path.endswith('.czi'):
        from eubi_bridge.base.czi_reader import read_czi as reader
        reader_kwargs = dict(
            as_mosaic = False,
            view_index = 0,
            phase_index = 0,
            illumination_index = 0,
            scene_index = 0,
            rotation_index = 0,
            mosaic_tile_index = 0,
            sample_index = 0
        )
        for k, v in reader_kwargs.items():
            if k in kwargs:
                reader_kwargs[k] = kwargs[k]
    elif input_path.endswith('.lif'):
        from bioio_lif.reader import Reader as reader
    elif input_path.endswith('.nd2'):
        from bioio_nd2.reader import Reader as reader
    elif input_path.endswith(('.png','.jpg','.jpeg')):
        from bioio_imageio.reader import Reader as reader
    else:
        from bioio_bioformats.reader import Reader as reader
    verbose = kwargs.get('verbose', False)
    if verbose:
        logger.info(f"Reading with {reader.__qualname__}.")
    im = reader(input_path, **reader_kwargs)
    if isinstance(im, da.Array):
        assert im.ndim == 5
        return im
    if hasattr(im, 'set_scene'):
        im.set_scene(kwargs.get('scene_index', 0))
        arr = im.get_image_dask_data('TCZYX')
    elif hasattr(im, 'base_array'):
        arr = im.base_array
    else:
        raise Exception(f"Unknown reader: {reader.__qualname__}")
    if arr.ndim > 5:
        arr = arr[0].transpose((0, 4, 1, 2, 3))
    return arr

def get_metadata_reader_by_path(input_path, **kwargs):
    if input_path.endswith(('ome.tiff', 'ome.tif')):
        from bioio_ome_tiff.reader import Reader as reader # pip install bioio-ome-tiff --no-deps
    elif input_path.endswith(('.tif', '.tiff', '.lsm')):
        from bioio_tifffile.reader import Reader as reader
    elif input_path.endswith('.czi'):
        from bioio_czi.reader import Reader as reader
    elif input_path.endswith('.lif'):
        from bioio_lif.reader import Reader as reader
    elif input_path.endswith('.nd2'):
        from bioio_nd2.reader import Reader as reader
    elif input_path.endswith(('.png','.jpg','.jpeg')):
        from bioio_imageio.reader import Reader as reader
    else:
        from bioio_bioformats.reader import Reader as reader
    return reader

def read_metadata_via_bioio_bioformats(input_path, **kwargs):
    from bioio_bioformats.reader import Reader
    series = kwargs.get('series', None)
    try:
        img = Reader(input_path)
        if series is not None:
            img.set_scene(series)
        omemeta = img.ome_metadata
    except FileNotFoundError as e:
        if ".jgo" in str(e):
            raise RuntimeError("JGO cache may be corrupted. Run `rm -rf ~/.jgo/` and retry.") from e
    return omemeta

def read_metadata_via_bfio(input_path, **kwargs):
    from bfio import BioReader
    try:
        omemeta = BioReader(input_path, backend='bioformats').metadata
    except FileNotFoundError as e:
        if ".jgo" in str(e):
            raise RuntimeError("JGO cache may be corrupted. Run `rm -rf ~/.jgo/` and retry.") from e
    return omemeta

def read_metadata_via_extension(input_path, **kwargs):
    Reader = get_metadata_reader_by_path(input_path)
    series = kwargs.get('series', None)
    try:
        img = Reader(input_path)
        if series is not None:
            img.set_scene(series)
        omemeta = img.ome_metadata
    except FileNotFoundError as e:
        if ".jgo" in str(e):
            raise RuntimeError("JGO cache may be corrupted. Run `rm -rf ~/.jgo/` and retry.") from e
    return omemeta



import fire
from eubi_bridge.ebridge import EuBIBridge
import multiprocessing as mp
import sys


def main():
    """
    Main function for the CLI interface.

    If the platform is Windows, sets the multiprocessing start method to "spawn".
    This is necessary because the default start method "fork" is not supported on Windows.
    """
    if sys.platform == "win32":
        mp.set_start_method("spawn", force=True)
    # Fire is a library for automatically generating CLIs from Python objects.
    # It uses the object's methods and docstrings to generate the CLI.
    fire.Fire(EuBIBridge)


if __name__ == "__main__":
    main()



import zarr, natsort

import shutil, time, os, zarr, pprint, psutil, dask, gc
import numpy as np, os, glob, tempfile, importlib

from ome_types.model import OME, Image, Pixels, Channel #TiffData, Plane
from ome_types.model import PixelType, Pixels_DimensionOrder, UnitsLength, UnitsTime

from typing import Tuple

from dask import array as da
from pathlib import Path
from typing import Union

from eubi_bridge.ngff.multiscales import Pyramid
from eubi_bridge.ngff.defaults import unit_map, scale_map, default_axes
from eubi_bridge.utils.convenience import sensitive_glob, is_zarr_group, is_zarr_array, take_filepaths
from eubi_bridge.base.readers import read_metadata_via_bioio_bioformats, read_metadata_via_extension, read_metadata_via_bfio


def abbreviate_units(measure: str) -> str:
    """Abbreviate a unit of measurement.

    Given a human-readable unit of measurement, return its abbreviated form.

    Parameters
    ----------
    measure : str
        The human-readable unit of measurement to abbreviate, e.g. "millimeter".

    Returns
    -------
    str
        The abbreviated form of the unit of measurement, e.g. "mm".

    Notes
    -----
    The abbreviations are as follows:

    * Length measurements:
        - millimeter: mm
        - centimeter: cm
        - decimeter: dm
        - meter: m
        - decameter: dam
        - hectometer: hm
        - kilometer: km
        - micrometer: µm
        - nanometer: nm
        - picometer: pm
    * Time measurements:
        - second: s
        - millisecond: ms
        - microsecond: µs
        - nanosecond: ns
        - minute: min
        - hour: h
    """
    if measure is None:
        return None

    abbreviations = {
        # Length measurements
        "millimeter": "mm",
        "centimeter": "cm",
        "decimeter": "dm",
        "meter": "m",
        "decameter": "dam",
        "hectometer": "hm",
        "kilometer": "km",
        "micrometer": "µm",
        "nanometer": "nm",
        "picometer": "pm",

        # Time measurements
        "second": "s",
        "millisecond": "ms",
        "microsecond": "µs",
        "nanosecond": "ns",
        "minute": "min",
        "hour": "h"
    }

    # Return the input if it's already an abbreviation
    if measure.lower() in abbreviations.values():
        return measure.lower()

    return abbreviations.get(measure.lower(), "Unknown")


def expand_units(measure: str) -> str:
    """
    Expand a unit of measurement.

    Given an abbreviated unit of measurement, return its expanded form.

    Parameters
    ----------
    measure : str
        The abbreviated unit of measurement to expand, e.g. "mm".

    Returns
    -------
    str
        The expanded form of the unit of measurement, e.g. "millimeter".
    """
    # Define the abbreviations and their expansions

    if measure is None:
        return None

    expansions = {
        # Length measurements
        "mm": "millimeter",
        "cm": "centimeter",
        "dm": "decimeter",
        "m": "meter",
        "dam": "decameter",
        "hm": "hectometer",
        "km": "kilometer",
        "µm": "micrometer",
        "nm": "nanometer",
        "pm": "picometer",

        # Time measurements
        "s": "second",
        "ms": "millisecond",
        "µs": "microsecond",
        "ns": "nanosecond",
        "min": "minute",
        "h": "hour"
    }

    # Return the input if it's already an expanded form
    if measure.lower() in expansions.values():
        return measure.lower()

    # Return the expanded form if it exists, else return "Unknown"
    return expansions.get(measure.lower(), "Unknown")

def create_ome_xml( # make 5D omexml
    image_shape: tuple,
    axis_order: str,
    pixel_size_x: float = None,
    pixel_size_y: float = None,
    pixel_size_z: float = None,
    pixel_size_t: float = None,
    unit_x: str = "MICROMETER",
    unit_y: str = None,
    unit_z: str = None,
    unit_t: str = None,
    dtype: str = "uint8",
    image_name: str = "Default Image",
    channel_names: list = None
    ) -> str:
    fullaxes = 'xyczt'
    if len(axis_order) != len(image_shape):
        raise ValueError("Length of axis_order must match length of image_shape")
    axis_order = axis_order.upper()

    pixel_size_basemap = {
        'time_increment': pixel_size_t,
        'physical_size_z': pixel_size_z,
        'physical_size_y': pixel_size_y,
        'physical_size_x': pixel_size_x
    }

    pixel_size_map = {}
    for ax in 'tzyx':
        if ax == 't':
            if ax in axis_order.lower():
                pixel_size_map['time_increment'] = pixel_size_t or 1
        else:
            if ax in axis_order.lower():
                pixel_size_map[f'physical_size_{ax}'] = pixel_size_basemap[f'physical_size_{ax}'] or 1


    unit_basemap = {
        'time_increment_unit': unit_t,
        'physical_size_z_unit': unit_z,
        'physical_size_y_unit': unit_y,
        'physical_size_x_unit': unit_x,
    }

    unit_map = {}
    for ax in 'tzyx':
        if ax == 't':
            if ax in axis_order.lower():
                unit_map['time_increment_unit'] = unit_t or 'second'
        else:
            if ax in axis_order.lower():
                unit_map[f'physical_size_{ax}_unit'] = unit_basemap[f'physical_size_{ax}_unit'] or 'MICROMETER'
    unit_map = {key: abbreviate_units(value) for key, value in unit_map.items() if value is not None}

    # Map numpy dtype to OME PixelType
    dtype_map = {
        "uint8": PixelType.UINT8,
        "uint16": PixelType.UINT16,
        "uint32": PixelType.UINT32,
        "int8": PixelType.INT8,
        "int16": PixelType.INT16,
        "int32": PixelType.INT32,
        "float": PixelType.FLOAT,
        "double": PixelType.DOUBLE,
    }

    if dtype not in dtype_map:
        raise ValueError(f"Unsupported dtype: {dtype}")

    pixel_type = dtype_map[dtype]

    # Initialize axis sizes
    size_map_ = dict(zip(axis_order.lower(), image_shape))
    size_map = {}
    for ax in fullaxes:
        if ax in size_map_:
            size_map[f'size_{ax}'] = size_map_[ax]
        else:
            size_map[f'size_{ax}'] = 1

    if channel_names is None or len(channel_names) != size_map['size_c']:
        channels=[Channel(id=f"Channel:{idx}", # TODO: if exists, directly take the channel names
                          samples_per_pixel=1)
                  for idx in range(size_map['size_c'])]
    else:
        channels=[Channel(id=f"Channel:{idx}", # TODO: if exists, directly take the channel names
                          samples_per_pixel=1,
                          name=channel_names[idx])
                  for idx in range(size_map['size_c'])]

    pixels = Pixels(
        dimension_order=Pixels_DimensionOrder(fullaxes.upper()),
        **size_map,
        type=pixel_type,
        **pixel_size_map,
        **unit_map,
        channels=channels,
    )

    image = Image(id="Image:0", name=image_name, pixels=pixels)

    ome = OME(images=[image])

    return ome


class PFFImageMeta:
    essential_omexml_fields = {
        "physical_size_x", "physical_size_x_unit",
        "physical_size_y", "physical_size_y_unit",
        "physical_size_z", "physical_size_z_unit",
        "time_increment", "time_increment_unit",
        "size_x", "size_y", "size_z", "size_t", "size_c"
    }
    def __init__(self,
                 path,
                 series,
                 meta_reader = "bioio"
                 ):
        if path.endswith('ome') or path.endswith('xml'):
            from ome_types import OME
            omemeta = OME().from_xml(path)
        else:
            if meta_reader == 'bioio':
                # Try to read the metadata via bioio
                try:
                    omemeta = read_metadata_via_extension(path, series = series)
                except:
                    # If not found, try to read the metadata via bioformats
                    omemeta = read_metadata_via_bioio_bioformats(path, series = series)
            elif meta_reader == 'bfio':
                try:
                    omemeta = read_metadata_via_bfio(path) # don't pass series, will be handled afterwards
                except:
                    # If not found, try to read the metadata via bioformats
                    omemeta = read_metadata_via_bioio_bioformats(path, series = series)
            else:
                raise ValueError(f"Unsupported metadata reader: {meta_reader}")
        if series is None:
            series = 0
        images = [omemeta.images[series]]
        omemeta.images = images
        self.omemeta = omemeta
        self.pixels = self.omemeta.images[0].pixels
        missing_fields = self.essential_omexml_fields - self.pixels.model_fields_set
        self.pixels.model_fields_set.update(missing_fields)
        self.omemeta.images[0].pixels = self.pixels
        self.pyr = None
        # self._channels = None

    def get_axes(self):
        return 'tczyx'

    def get_scaledict(self):
        return {
            't': self.pixels.time_increment,
            'z': self.pixels.physical_size_z,
            'y': self.pixels.physical_size_y,
            'x': self.pixels.physical_size_x
        }

    def get_scales(self):
        scaledict = self.get_scaledict()
        caxes = [ax for ax in self.get_axes() if ax != 'c']
        return [scaledict[ax] for ax in caxes]

    def get_unitdict(self):
        return {
            't': self.pixels.time_increment_unit.name.lower(),
            'z': self.pixels.physical_size_z_unit.name.lower(),
            'y': self.pixels.physical_size_y_unit.name.lower(),
            'x': self.pixels.physical_size_x_unit.name.lower()
        }

    def get_units(self):
        unitdict = self.get_unitdict()
        caxes = [ax for ax in self.get_axes() if ax != 'c']
        return [unitdict[ax] for ax in caxes]

    def get_channels(self):
        if not hasattr(self.pixels, 'channels'):
            return None
        if len(self.pixels.channels) == 0:
            return None
        ###
        if len(self.pixels.channels) < self.pixels.size_c:
            chn = ChannelIterator(num_channels = self.pixels.size_c)
            channels = chn._channels
        elif len(self.pixels.channels) == self.pixels.size_c:
            channels = []
            for _, channel in enumerate(self.pixels.channels):
                color = channel.color.as_hex().upper()
                color = expand_hex_shorthand(color)
                name = channel.name
                channels.append(dict(
                    label = name,
                    color = color
                ))
        return channels

def expand_hex_shorthand(hex_color):
    """
    Expands a shorthand hex color of any valid length (e.g., #abc → #aabbcc, #1234 → #11223344).
    """
    if not hex_color.startswith('#'):
        raise ValueError("Hex color must start with '#'")

    shorthand = hex_color[1:]

    if not all(c in '0123456789ABCDEFabcdef' for c in shorthand):
        raise ValueError("Invalid hex digits")

    expanded = '#' + ''.join([c * 2 for c in shorthand])
    return expanded

class NGFFImageMeta:
    def __init__(self,
                 path
                 ):
        if is_zarr_group(path):
            self.pyr = Pyramid().from_ngff(path)
            meta = self.pyr.meta
            self._meta = meta
            self._base_path = self._meta.resolution_paths[0]
        else:
            raise Exception(f"The given path does not contain an NGFF group.")

    def get_axes(self):
        return self._meta.axis_order

    def get_scales(self):
        return self._meta.get_scale(self._base_path)

    def get_scaledict(self):
        return self._meta.get_scaledict(self._base_path)

    def get_units(self):
        return self._meta.unit_list

    def get_unitdict(self):
        return self._meta.unit_dict

    def get_channels(self):
        if not hasattr(self._meta, 'channels'):
            return None
        return self._meta.channels


class ArrayManager:
    essential_omexml_fields = {
        "physical_size_x", "physical_size_x_unit",
        "physical_size_y", "physical_size_y_unit",
        "physical_size_z", "physical_size_z_unit",
        "time_increment", "time_increment_unit",
        "size_x", "size_y", "size_z", "size_t", "size_c"
    }
    def __init__(self,
                 path: Union[str, Path] = None,
                 series: int = None,
                 metadata_reader='bfio',  # bfio or aicsimageio
                 **kwargs
                 ):
        self.path = path
        self.series = series
        if series is not None:
            assert isinstance(self.series, (int,
                                            str)), f"The series parameter must be either an integer or string. Selection of multiple series from the same image is currently not supported."
        if self.series is None:
            self.series = 0
            self._seriesattrs = ""
        else:
            self._seriesattrs = self.series

        self._meta_reader = metadata_reader
        self.omemeta = None

        if not path is None:
            if is_zarr_group(path):
                self.img = NGFFImageMeta(self.path)
            else:
                self.img = PFFImageMeta(self.path, self.series, self._meta_reader)

        self.axes = self.img.get_axes()
        self.array = None
        self.pyr = self.img.pyr
        self._channels = None
        self.set_arraydata()

    def fill_default_meta(self):
        if self.array is None:
            raise Exception(f"Array is missing. An array needs to be assigned.")
        new_scaledict = {}
        new_unitdict = {}
        values = list(self.scaledict.values())
        if not None in values:
            return

        for ax, value in self.scaledict.items():
            if value is None:
                if (ax == 'z' or ax == 'y') and self.scaledict['x'] is not None:
                    new_scaledict[ax] = self.scaledict['x']
                    new_unitdict[ax] = self.unitdict['x']
                else:
                    new_scaledict[ax] = scale_map[ax]
                    new_unitdict[ax] = unit_map[ax]
            else:
                if ax in self.scaledict.keys():
                    new_scaledict[ax] = self.scaledict[ax]
                if ax in self.unitdict.keys():
                    new_unitdict[ax] = self.unitdict[ax]

        new_units = [new_unitdict[ax] for ax in self.axes if ax in new_unitdict]
        new_scales = [new_scaledict[ax] for ax in self.axes if ax in new_scaledict]

        self.set_arraydata(self.array, self.axes, new_units, new_scales)
        return self

    def get_pixel_size_basemap(self,
                               t = 1,
                               z = 1,
                               y = 1,
                               x = 1,
                               **kwargs
                               ):
        return {
            'pixel_size_t': t,
            'pixel_size_z': z,
            'pixel_size_y': y,
            'pixel_size_x': x
        }

    def get_unit_basemap(self,
                         t = 'second',
                         z = 'micrometer',
                         y = 'micrometer',
                         x = 'micrometer',
                         **kwargs
                         ):
        return {
            'unit_t': t,
            'unit_z': z,
            'unit_y': y,
            'unit_x': x
        }

    def update_meta(self,
                    new_scaledict = {},
                    new_unitdict = {}
                    ):

        scaledict = self.img.get_scaledict()
        for key, val in new_scaledict.items():
            if key in scaledict.keys() and val is not None:
                scaledict[key] = val

        if 'c' in scaledict:
            scales = [scaledict[ax] for ax in self.axes]
        else:
            scales = [scaledict[ax] for ax in self.caxes]

        unitdict = self.img.get_unitdict()
        for key, val in new_unitdict.items():
            if key in unitdict.keys() and val is not None:
                unitdict[key] = val

        if 'c' in unitdict:
            units = [expand_units(unitdict[ax]) for ax in self.axes]
        else:
            units = [expand_units(unitdict[ax]) for ax in self.caxes]

        self.set_arraydata(array = self.array,
                           axes = self.axes,
                           units = units,
                           scales = scales)

    def _ensure_correct_channels(self):
        if self.array is None:
            return
        if self.channels is None:
            return
        shapedict = dict(zip(list(self.axes), self.array.shape))
        csize = shapedict['c']
        channelsize = len(self.channels)

        if channelsize > csize:
            self._channels = [channel for channel in self.channels
                        if channel['label'] is not None]

    def fix_bad_channels(self):
        ### Update channel labels
        chn = ChannelIterator()
        for i, channel in enumerate(self.channels):
            if channel['label'] is None:
                channel = chn.__next__()
            self.channels[i] = channel

    def set_arraydata(self,
                      array = None,
                      axes = None,
                      units = None,
                      scales = None,
                      **kwargs # placehold
                      ):

        axes = axes or self.img.get_axes()
        units = units or self.img.get_units()
        scales = scales or self.img.get_scales()

        self.axes = axes
        if array is not None:
            self.array = array
            self.ndim = self.array.ndim
            assert len(self.axes) == self.ndim

        self.caxes = ''.join([ax for ax in axes if ax != 'c'])
        if self.array is not None:
            self.chunkdict = dict(zip(list(self.axes), self.array.chunksize))
            self.shapedict = dict(zip(list(self.axes), self.array.shape))
            if 'c' in self.shapedict:
                self._ensure_correct_channels()

        if len(units) == len(self.axes):
            self.unitdict = dict(zip(list(self.axes), units))
        elif len(units) == len(self.caxes):
            self.unitdict = dict(zip(list(self.caxes), units))
        else:
            raise Exception(f"Unit length is invalid.")
        if len(scales) == len(self.axes):
            self.scaledict = dict(zip(list(self.axes), scales))
        elif len(scales) == len(self.caxes):
            self.scaledict = dict(zip(list(self.caxes), scales))
            self.scaledict['c'] = 1
        else:
            raise Exception(f"Scale length is invalid")

    @property
    def scales(self):
        if self.scaledict.__len__() < len(self.axes):
            return [self.scaledict[ax] for ax in self.caxes]
        elif self.scaledict.__len__() == len(self.axes):
            return [self.scaledict[ax] for ax in self.axes]
        else:
            raise ValueError

    @property
    def units(self):
        if self.unitdict.__len__() < len(self.axes):
            return [self.unitdict[ax] for ax in self.caxes]
        elif self.unitdict.__len__() == len(self.axes):
            return [self.unitdict[ax] for ax in self.axes]
        else:
            raise ValueError
    
    @property
    def channels(self):
        if self._channels is not None:
            return self._channels
        return self.img.get_channels()

    @property
    def chunks(self):
        return [self.chunkdict[ax] for ax in self.axes]

    def sync_pyramid(self,
                     create_omexml_if_not_exists = False
                     ): ### TODO: reconsider for improvement
        ### This is only to be used to update pyramidal metadata in place.
        """
        Synchronizes the scale and unit metadata with the Pyramid (if a Pyramid exists).
        The scale metadata is recalculated for all layers based on the shape of each layer.
        Also updates the ome-xml metadata to the pyramid.
        :return:
        """
        if self.pyr is None:
            raise Exception(f"No pyramid exists.")

        self.pyr.update_scales(**self.scaledict,
                                     # hard=True
                               )
        self.pyr.update_units(**self.unitdict,
                                   # hard=True
                              )
        if self.omemeta is None:
            self.omemeta = create_ome_xml(
                image_shape=self.pyr.base_array.shape,
                axis_order=self.pyr.axes,
                pixel_size_x=self.pyr.meta.scaledict.get('0', {}).get('x'),
                pixel_size_y=self.pyr.meta.scaledict.get('0', {}).get('y'),
                pixel_size_z=self.pyr.meta.scaledict.get('0', {}).get('z'),
                pixel_size_t=self.pyr.meta.scaledict.get('0', {}).get('t'),
                unit_x=self.pyr.meta.unit_dict.get('x'),
                unit_y=self.pyr.meta.unit_dict.get('y'),
                unit_z=self.pyr.meta.unit_dict.get('z'),
                unit_t=self.pyr.meta.unit_dict.get('t'),
                dtype=str(self.pyr.base_array.dtype),
                image_name=self.pyr.meta.multiscales.get('name', 'Default image'),
                channel_names=[channel['label'] for channel in self.channels]
            )
        if 'OME' in list(self.pyr.gr.keys()) or create_omexml_if_not_exists:
            # If OME-XML exists in the pyramid, update it.
            # Otherwise create a new ome-xml only if create_omexml_if_not_exists is True
            # Otherwise do not create a new ome-xml
            self.save_omexml(self.pyr.gr.store.root, overwrite=True)
        self.pyr.meta.save_changes()

    def create_omemeta(self):
        self.fill_default_meta()

        pixel_size_basemap = self.get_pixel_size_basemap(
            **self.scaledict
        )
        unit_basemap = self.get_unit_basemap(
            **self.unitdict
        )
        self.omemeta = create_ome_xml(image_shape = self.array.shape,
                                      axis_order = self.axes,
                                      **pixel_size_basemap,
                                      **unit_basemap,
                                      dtype = str(self.array.dtype),
                                      channel_names = [channel['label'] for channel in self.channels]
                                      )
        self.pixels = self.omemeta.images[0].pixels
        missing_fields = self.essential_omexml_fields - self.pixels.model_fields_set
        self.pixels.model_fields_set.update(missing_fields)
        self.omemeta.images[0].pixels = self.pixels
        return self

    def save_omexml(self,
                    base_path: str,
                    overwrite = False
                    ):
        assert self.omemeta is not None, f"No ome-xml exists."
        gr = zarr.group(base_path)
        gr.create_group('OME', overwrite = overwrite)

        path = os.path.join(gr.store.root, 'OME/METADATA.ome.xml')

        with open(path, 'w', encoding='utf-8') as f:
            f.write(self.omemeta.to_xml())

        if gr.info._zarr_format == 2:
            gr['OME'].attrs["series"] = [self._seriesattrs]
        else: # zarr format 3
            gr['OME'].attrs["ome"] = dict(version = "0.5",
                                          series = [str(self._seriesattrs)]
                                          )

    def squeeze(self):
        singlet_axes = [ax for ax, size in self.shapedict.items() if size == 1]
        newaxes = ''.join(ax for ax in self.axes if ax not in singlet_axes)
        newunits, newscales = [], []
        assert (self.scaledict.__len__() - self.unitdict.__len__()) <= 1
        for ax in self.axes:
            if ax not in singlet_axes:
                if ax in self.unitdict.keys():
                    newunits.append(self.unitdict[ax])
                if ax in self.scaledict.keys():
                    newscales.append(self.scaledict[ax])
        newarray = da.squeeze(self.array)
        self.set_arraydata(newarray, newaxes, newunits, newscales)

    def transpose(self, newaxes):
        newaxes = ''.join(ax for ax in newaxes if ax in self.axes)
        new_ids = [self.axes.index(ax) for ax in newaxes]
        newunits, newscales = [], []
        assert (self.scaledict.__len__() - self.unitdict.__len__()) <= 1
        for ax in newaxes:
            if ax in self.unitdict:
                newunits.append(self.unitdict[ax])
            if ax in self.scaledict.keys():
                newscales.append(self.scaledict[ax])
        newarray = self.array.transpose(*new_ids)
        self.set_arraydata(newarray, newaxes, newunits, newscales)

    def crop(self,
             trange = None,
             crange = None,
             zrange = None,
             yrange = None,
             xrange = None,
             ):
        slicedict = {
            't': slice(*trange) if trange is not None else slice(None),
            'c': slice(*crange) if crange is not None else slice(None),
            'z': slice(*zrange) if zrange is not None else slice(None),
            'y': slice(*yrange) if yrange is not None else slice(None),
            'x': slice(*xrange) if xrange is not None else slice(None),
        }
        slicedict = {ax: r for ax,r in slicedict.items() if ax in self.axes}
        slices = tuple([slicedict[ax] for ax in self.axes])
        array = self.array[slices]
        self.set_arraydata(array, self.axes, self.units, self.scales)

    def to_cupy(self):
        try:
            import cupy
        except:
            raise Exception("Cupy not installed but required for this operation.")
        array = self.array.map_blocks(cupy.asarray)
        self.set_arraydata(array, self.axes, self.units, self.scales)

    def split(self): ###TODO
        pass

class ChannelIterator:
    """
    Iterator for generating and managing channel colors.

    This class provides a way to iterate through a sequence of channel colors,
    generating new colors in a visually distinct sequence when needed.
    """
    DEFAULT_COLORS = [
        "FF0000",  # Red
        "00FF00",  # Green
        "0000FF",  # Blue
        "FF00FF",  # Magenta
        "00FFFF",  # Cyan
        "FFFF00",  # Yellow
        "FFFFFF",  # White
    ]

    def __init__(self, num_channels=0):
        """
        Initialize the channel iterator.

        Args:
            num_channels: Initial number of channels to pre-generate
        """
        self._channels = []
        self._current_index = 0
        self._generate_channels(num_channels)

    def _generate_channels(self, count):
        """Generate the specified number of unique channel colors."""
        for i in range(len(self._channels), count):
            if i < len(self.DEFAULT_COLORS):
                color = self.DEFAULT_COLORS[i]
            else:
                # Generate a distinct color by distributing hues
                hue = int((i * 137.5) % 360)  # Golden angle for distinct colors
                r, g, b = self._hsv_to_rgb(hue / 360.0, 1.0, 1.0)
                color = f"{int(r * 255):02X}{int(g * 255):02X}{int(b * 255):02X}"
            self._channels.append({"label": f"Channel {i + 1}", "color": color})

    @staticmethod
    def _hsv_to_rgb(h, s, v):
        """Convert HSV color space to RGB color space."""
        h = h * 6.0
        i = int(h)
        f = h - i
        p = v * (1.0 - s)
        q = v * (1.0 - s * f)
        t = v * (1.0 - s * (1.0 - f))

        if i == 0:
            return v, t, p
        elif i == 1:
            return q, v, p
        elif i == 2:
            return p, v, t
        elif i == 3:
            return p, q, v
        elif i == 4:
            return t, p, v
        else:
            return v, p, q

    def __iter__(self):
        """Return the iterator object itself."""
        self._current_index = 0
        return self

    def __next__(self):
        """Return the next channel color."""
        if self._current_index >= len(self._channels):
            self._generate_channels(len(self._channels) + 1)

        if self._current_index < len(self._channels):
            result = self._channels[self._current_index]
            self._current_index += 1
            return result
        raise StopIteration

    def get_channel(self, index):
        """
        Get channel color by index.

        Args:
            index: Index of the channel to retrieve

        Returns:
            dict: Channel information with 'label' and 'color' keys
        """
        if index >= len(self._channels):
            self._generate_channels(index + 1)
        return self._channels[index]

    def __len__(self):
        """Return the number of generated channels."""
        return len(self._channels)


class BatchManager:
    def __init__(self,
                 managers
                 ):
        self.managers = managers

    def _collect_scaledict(self, **kwargs):
        """
        Retrieves pixel sizes for image dimensions.

        Args:
            **kwargs: Pixel sizes for time, channel, z, y, and x dimensions.

        Returns:
            list: Pixel sizes.
        """
        t = kwargs.get('time_scale', None)
        c = kwargs.get('channel_scale', None)
        y = kwargs.get('y_scale', None)
        x = kwargs.get('x_scale', None)
        z = kwargs.get('z_scale', None)
        fulldict = dict(zip('tczyx', [t, c, z, y, x]))
        final = {key: val for key, val in fulldict.items() if val is not None}
        return final

    def _collect_unitdict(self, **kwargs):
        """
        Retrieves unit specifications for image dimensions.

        Args:
            **kwargs: Unit values for time, channel, z, y, and x dimensions.

        Returns:
            list: Unit values.
        """
        t = kwargs.get('time_unit', None)
        c = kwargs.get('channel_unit', None)
        y = kwargs.get('y_unit', None)
        x = kwargs.get('x_unit', None)
        z = kwargs.get('z_unit', None)
        fulldict = dict(zip('tczyx', [t, c, z, y, x]))
        final = {key: val for key, val in fulldict.items() if val is not None}
        return final

    def _collect_chunks(self, **kwargs):  ###
        """
        Retrieves chunk specifications for image dimensions.

        Args:
            **kwargs: Chunk sizes for time, channel, z, y, and x dimensions.

        Returns:
            list: Chunk shape.
        """
        t = kwargs.get('time_chunk', None)
        c = kwargs.get('channel_chunk', None)
        y = kwargs.get('y_chunk', None)
        x = kwargs.get('x_chunk', None)
        z = kwargs.get('z_chunk', None)
        fulldict = dict(zip('tczyx', [t, c, z, y, x]))
        final = {key: val for key, val in fulldict.items() if val is not None}
        return final

    def fill_default_meta(self):
        for key, manager in self.managers.items():
            manager.fill_default_meta()

    def squeeze(self):
        for key, manager in self.managers.items():
            manager.squeeze()

    def to_cupy(self):
        for key, manager in self.managers.items():
            manager.to_cupy()

    def crop(self,
             time_range=None,
             channel_range=None,
             z_range=None,
             y_range=None,
             x_range=None,
             **kwargs  # placehold
             ):
        if any([item is not None
                for item in (time_range,
                             channel_range,
                             z_range,
                             y_range,
                             x_range)]):
            for key, manager in self.managers.items():
                manager.crop(time_range, channel_range, z_range, y_range, x_range)

    def transpose(self, newaxes):
        for key, manager in self.managers.items():
            manager.transpose(newaxes)

    def sync_pyramids(self):
        pass


import copy
import os, itertools, tempfile, shutil, threading
import pprint
import zarr, dask, numcodecs
from zarr import codecs
from zarr.storage import LocalStore
from dataclasses import dataclass
from dask import delayed
import dask.array as da
import numpy as np
from pathlib import Path
from typing import List, Tuple, Dict, Union, Any, Tuple, Optional
### internal imports
from eubi_bridge.ngff.multiscales import NGFFMetadataHandler  #Multimeta
from eubi_bridge.utils.convenience import get_chunksize_from_array, is_zarr_group #, retry_decorator


import logging, warnings

logging.getLogger('distributed.diskutils').setLevel(logging.CRITICAL)

ZARR_V2 = 2
ZARR_V3 = 3
DEFAULT_DIMENSION_SEPARATOR = "/"
DEFAULT_COMPRESSION_LEVEL = 5
DEFAULT_COMPRESSION_ALGORITHM = "zstd"


@dataclass
class CompressorConfig:
    name: str = 'blosc'
    params: dict = None

    def __post_init__(self):
        self.params = self.params or {}

def autocompute_color(channel_ix: int):
    default_colors = [
        "FF0000",  # Red
        "00FF00",  # Green
        "0000FF",  # Blue
        "FF00FF",  # Magenta
        "00FFFF",  # Cyan
        "FFFF00",  # Yellow
        "FFFFFF",  # White
    ]
    color = default_colors[i] if i < len(default_colors) else f"{i * 40 % 256:02X}{i * 85 % 256:02X}{i * 130 % 256:02X}"
    return color

def create_zarr_array(directory: Union[Path, str, zarr.Group],
                      array_name: str,
                      shape: Tuple[int, ...],
                      chunks: Tuple[int, ...],
                      dtype: Any,
                      overwrite: bool = False) -> zarr.Array:
    chunks = tuple(np.minimum(shape, chunks))

    if not isinstance(directory, zarr.Group):
        path = os.path.join(directory, array_name)
        dataset = zarr.create(shape=shape,
                              chunks=chunks,
                              dtype=dtype,
                              store=path,
                              dimension_separator='/',
                              overwrite=overwrite)
    else:
        dataset = directory.create(name=array_name,
                                   shape=shape,
                                   chunks=chunks,
                                   dtype=dtype,
                                   dimension_separator='/',
                                   overwrite=overwrite)
    return dataset


def get_regions(array_shape: Tuple[int, ...],
                region_shape: Tuple[int, ...],
                as_slices: bool = False) -> list:
    assert len(array_shape) == len(region_shape)
    steps = []
    for size, inc in zip(array_shape, region_shape):
        seq = np.arange(0, size, inc)
        if size > seq[-1]:
            seq = np.append(seq, size)
        increments = tuple((seq[i], seq[i + 1]) for i in range(len(seq) - 1))
        if as_slices:
            steps.append(tuple(slice(*item) for item in increments))
        else:
            steps.append(increments)
    return list(itertools.product(*steps))


def get_compressor(name,
                   zarr_format = ZARR_V2,
                   **params): ### TODO: continue this, add for zarr3
    name = name.lower()
    assert zarr_format in (ZARR_V2, ZARR_V3)
    compression_dict2 = {
        "blosc": "Blosc",
        "bz2": "BZ2",
        "gzip": "GZip",
        "lzma": "LZMA",
        "lz4": "LZ4",
        "pcodec": "PCodec",
        "zfpy": "ZFPY",
        "zlib": "Zlib",
        "zstd": "Zstd"
    }

    compression_dict3 = {
        "blosc": "BloscCodec",
        "gzip": "GzipCodec",
        "sharding": "ShardingCodec",
        "zstd": "ZstdCodec",
        "crc32ccodec": "CRC32CCodec"
    }

    if zarr_format == ZARR_V2:
        compressor_name = compression_dict2[name]
        compressor_instance = getattr(numcodecs, compressor_name)
    elif zarr_format == ZARR_V3:
        compressor_name = compression_dict3[name]
        compressor_instance = getattr(codecs, compressor_name)
    else:
        raise Exception("Unsupported Zarr format")
    compressor = compressor_instance(**params)
    return compressor

def get_default_fill_value(dtype):
    if np.issubdtype(dtype, np.integer):
        return 0
    elif np.issubdtype(dtype, np.floating):
        return 0.0
    elif np.issubdtype(dtype, np.bool_):
        return False
    return None

def _create_zarr_v2_array(
        store_path: Union[Path, str],
        shape: Tuple[int, ...],
        chunks: Tuple[int, ...],
        dtype: Any,
        compressor_config: CompressorConfig,
        dimension_separator: str,
        overwrite: bool,
) -> zarr.Array:
    compressor = get_compressor(compressor_config.name,
                                zarr_format=ZARR_V2,
                                **compressor_config.params)
    return zarr.create(
        shape=shape,
        chunks=chunks,
        dtype=dtype,
        store=store_path,
        compressor=compressor,
        dimension_separator=dimension_separator,
        overwrite=overwrite,
        zarr_format=ZARR_V2,
    )

def _create_zarr_v3_array(
        store: Any,
        shape: Tuple[int, ...],
        chunks: Tuple[int, ...],
        dtype: Any,
        compressor_config: CompressorConfig,
        shards: Optional[Tuple[int, ...]],
        dimension_names: str = None,
        overwrite: bool = False,
        **kwargs
) -> zarr.Array:
    compressors = [get_compressor(compressor_config.name,
                                  zarr_format=ZARR_V3,
                                  **compressor_config.params)
                   ]
    return zarr.create_array(
        store=store,
        shape=shape,
        chunks=chunks,
        shards=shards,
        dimension_names=dimension_names,
        dtype=dtype,
        compressors=compressors,
        overwrite=overwrite,
        zarr_format=ZARR_V3,
        **kwargs
    )

def _create_zarr_array(
        store_path: Union[Path, str],
        shape: Tuple[int, ...],
        chunks: Tuple[int, ...],
        dtype: Any,
        compressor_config: CompressorConfig = None,
        zarr_format: int = ZARR_V2,
        overwrite: bool = False,
        shards: Optional[Tuple[int, ...]] = None,
        dimension_separator: str = DEFAULT_DIMENSION_SEPARATOR,
        dimension_names: str = None,
        **kwargs
) -> zarr.Array:
    """Create a Zarr array with specified format and compression settings."""
    compressor_config = compressor_config or CompressorConfig()
    chunks = tuple(np.minimum(shape, chunks).tolist())
    if shards is not None:
        shards = tuple(np.array(shards).flatten().tolist())
        assert np.allclose(np.mod(shards, chunks), 0)
    store = LocalStore(store_path)

    if zarr_format not in (ZARR_V2, ZARR_V3):
        raise ValueError(f"Unsupported Zarr format: {zarr_format}")

    if zarr_format == ZARR_V2:
        return _create_zarr_v2_array(
            store_path=store_path,
            shape=shape,
            chunks=chunks,
            dtype=dtype,
            compressor_config=compressor_config,
            dimension_separator=dimension_separator,
            overwrite=overwrite,
        )

    return _create_zarr_v3_array(
        store=store,
        shape=shape,
        chunks=chunks,
        dtype=dtype,
        compressor_config=compressor_config,
        shards=shards,
        dimension_names=dimension_names,
        overwrite=overwrite,
        # **kwargs
    )

def write_chunk_with_zarrpy(chunk: np.ndarray, zarr_array: zarr.Array, block_info: Dict) -> None:
    if hasattr(chunk, "get"):
        chunk = chunk.get()  # Convert CuPy -> NumPy
    zarr_array[tuple(slice(*b) for b in block_info[0]["array-location"])] = chunk

def write_with_zarrpy(arr: da.Array,
                      store_path: Union[str, Path],
                      chunks: Optional[Tuple[int, ...]] = None,
                      shards: Optional[Tuple[int, ...]] = None,
                      dimension_names: str = None,
                      dtype: Any = None,
                      compressor: str = 'blosc',
                      compressor_params: dict = None,
                      rechunk_method: str = 'tasks',
                      overwrite: bool = True,
                      zarr_format: int = 2,
                      **kwargs
                      ) -> zarr.Array:
    """
    Write dask array to zarr storage using either zarr v2 or v3 format.

    Args:
        arr: Dask array to write
        store_path: Path where the Zarr array will be stored
        chunks: Chunk size for each dimension
        shards: Shard size for zarr v3 format
        dtype: Data type of the array (defaults to arr.dtype)
        compressor: Compression algorithm ('blosc' by default)
        compressor_params: Parameters for the compressor
        rechunk_method: Method for rechunking ('tasks' or 'p2p')
        zarr_format: Zarr format version (2 or 3)
        **kwargs: Additional arguments for array creation
    """
    store_path = str(store_path)
    dtype = dtype or arr.dtype
    compressor_params = compressor_params or {'cname': 'zstd', 'clevel': 5}

    if chunks is None:
        chunks = arr.chunksize

    chunks = tuple(int(size) for size in chunks)

    if shards is None:
        shards = copy.deepcopy(chunks)

    if not np.allclose(np.mod(shards, chunks), 0):
        multiples = np.floor_divide(shards, chunks)
        shards = np.multiply(multiples, chunks)

    shards = tuple(int(size) for size in np.ravel(shards))

    if zarr_format == 2:
        if not np.equal(arr.chunksize, chunks).all():
            arr = arr.rechunk(chunks, method=rechunk_method)
    else:  # zarr_format == 3
        if shards is not None and not np.equal(arr.chunksize, shards).all():
            arr = arr.rechunk(shards, method=rechunk_method)

    compressor_config = CompressorConfig(name=compressor,
                                         params=compressor_params)

    zarr_array = _create_zarr_array(
        store_path=store_path,
        shape=arr.shape,
        chunks=chunks,
        dtype=dtype,
        overwrite=overwrite,
        compressor_config = compressor_config,
        zarr_format=zarr_format,
        shards=shards,
        dimension_names=dimension_names,
        **kwargs
    )
    res = arr.map_blocks(write_chunk_with_zarrpy, zarr_array=zarr_array, dtype=dtype)

    return res

def write_chunk_with_tensorstore(chunk: np.ndarray, ts_store, block_info: Dict) -> None:
    if hasattr(chunk, "get"):
        chunk = chunk.get()  # Convert CuPy -> NumPy
    ts_store[tuple(slice(*b) for b in block_info[0]["array-location"])] = chunk


def write_with_tensorstore(
    arr: da.Array,
    store_path: Union[str, Path],
    chunks: Optional[Tuple[int, ...]] = None,
    shards: Optional[Tuple[int, ...]] = None,
    dimension_names: str = None,
    dtype: Any = None,
    compressor: str = 'blosc',
    compressor_params: dict = None,
    rechunk_method: str = 'tasks',
    overwrite: bool = True,
    zarr_format: int = 2,
    **kwargs
) -> da.Array:
    """
    Write dask array to zarr storage using tensorstore with support for both zarr v2 and v3 formats.
    """
    try:
        import tensorstore as ts
    except ImportError:
        raise ModuleNotFoundError(
            "The module tensorstore has not been found. "
            "Try 'conda install -c conda-forge tensorstore'"
        )

    compressor_params = compressor_params or {'cname': 'zstd', 'clevel': 5}
    # shard_to_chunk_ratio = kwargs.get('shard_to_chunk_ratio', 3)
    dtype = dtype or arr.dtype
    fill_value = kwargs.get('fill_value', get_default_fill_value(dtype))

    if chunks is None:
        chunks = arr.chunksize

    chunks = tuple(int(size) for size in chunks)

    if shards is None:
        shards = copy.deepcopy(chunks)

    if not np.allclose(np.mod(shards, chunks), 0):
        multiples = np.floor_divide(shards, chunks)
        shards = np.multiply(multiples, chunks)

    shards = tuple(int(size) for size in np.ravel(shards))

    # Rechunk if needed
    if zarr_format == 2:
        if not np.equal(arr.chunksize, chunks).all():
            arr = arr.rechunk(chunks, method=rechunk_method)
    else:  # zarr_format == 3
        if shards is not None and not np.equal(arr.chunksize, shards).all():
            arr = arr.rechunk(shards, method=rechunk_method)

    # Prepare zarr metadata
    if zarr_format == 3:
        # Only include array-to-array codecs such as blosc
        zarr_metadata = {
            "data_type": np.dtype(dtype).name,
            "shape": arr.shape,
            "chunk_grid": {
                "name": "regular",
                "configuration": {
                    "chunk_shape": shards if zarr_format == 3 else chunks,
                }
            },
            "dimension_names": list(dimension_names),
            "codecs": [
                {
                    "name": "sharding_indexed",
                    "configuration": {
                        "chunk_shape": chunks,
                        "codecs": [{"name": "bytes",
                                    "configuration": {"endian": "little"}},
                                   {"name": compressor,
                                    "configuration": compressor_params or {}}],
                        "index_codecs": [{"name": "bytes",
                                          "configuration": {"endian": "little"}},
                                         {"name": "crc32c"}],
                        "index_location": "end"
                    }
                }
            ],
            "node_type": "array"
        }

    else:  # zarr_format == 2
        zarr_metadata = {
            "compressor": {
                "id": compressor,
                **compressor_params
            },
            "dtype": np.dtype(dtype).str,
            "shape": arr.shape,
            "chunks": chunks,
            "fill_value": fill_value,
            "dimension_separator": '/',
        }

    zarr_spec = {
        "driver": "zarr" if zarr_format == 2 else "zarr3",
        "kvstore": {
            "driver": "file",
            "path": str(store_path),
        },
        "metadata": zarr_metadata,
        "create": True,
        "delete_existing": overwrite,
    }

    ts_store = ts.open(zarr_spec).result()
    return arr.map_blocks(write_chunk_with_tensorstore, ts_store=ts_store, dtype=dtype)



@delayed
def count_threads():
    return threading.active_count()


def _get_or_create_multimeta(gr: zarr.Group,
                             axis_order: str,
                             unit_list: List[str],
                             version: str) -> NGFFMetadataHandler:
    """
    Read existing or create new metadata handler for zarr group.

    Parameters
    ----------
    gr : zarr.Group
        Zarr group to read metadata from or write metadata to.
    axis_order : str
        String indicating the order of axes in the arrays.
    unit_list : List[str]
        List of strings indicating the units of each axis.
    version : str
        Version of NGFF to create if no metadata exists.

    Returns
    -------
    handler : NGFFMetadataHandler
        Metadata handler for the zarr group.
    """
    handler = NGFFMetadataHandler()
    handler.connect_to_group(gr)
    try:
        handler.read_metadata()
    except:
        handler.create_new(version=version)
        handler.parse_axes(axis_order=axis_order, units=unit_list)
    return handler


def store_arrays(arrays: Dict[str, Dict[str, da.Array]], # flatarrays
                 output_path: Union[Path, str],
                 axes: list, # flataxes
                 scales: Dict[str, Dict[str, Tuple[float, ...]]], # flatscales
                 units: list, # flatunits
                 output_chunks: Tuple[int, ...] = None,
                 output_shard_coefficients: Tuple[int, ...] = None,
                 compute: bool = False,
                 overwrite: bool = False,
                 channel_meta: dict = None,
                 **kwargs # shard_to_chunk_ratio and zarr_format should specified inside kwargs
                 ) -> Dict[str, da.Array]:

    rechunk_method = kwargs.get('rechunk_method', 'tasks')
    if rechunk_method == 'rechunker':
        raise ValueError(f"This version of EuBI-Bridge does not support rechunker. Choose either of 'tasks' or 'p2p'.")
    assert rechunk_method in ('tasks', 'p2p')
    use_tensorstore = kwargs.get('use_tensorstore', False)
    verbose = kwargs.get('verbose', False)
    zarr_format = kwargs.get('zarr_format', 2)
    output_shards = kwargs.get('output_shards', None)

    writer_func = write_with_tensorstore if use_tensorstore else write_with_zarrpy

    zarr.group(output_path, overwrite=overwrite, zarr_version = zarr_format)
    results = {}

    for key, arr in arrays.items():
        flataxes = axes[key]
        flatscale = scales[key]
        flatunit = units[key]
        flatchunks = output_chunks[key]
        flatchannels = channel_meta[key] if channel_meta is not None else None

        chunks = np.minimum(flatchunks or arr.chunksize, arr.shape).tolist()
        chunks = tuple([int(item) for item in chunks])

        if zarr_format == 3:
            if output_shards is not None:
                shards = output_shards[key]
            elif output_shard_coefficients is not None:
                flatshardcoefs = output_shard_coefficients[key]
                shards = np.multiply(chunks, flatshardcoefs)
            else:
                shards = chunks

            shards = tuple([int(item) for item in shards])
        else:
            shards = None

        dirpath = os.path.dirname(key)
        arrpath = os.path.basename(key)

        if is_zarr_group(dirpath):
            gr = zarr.open_group(dirpath, mode='a')
        else:
            gr = zarr.group(dirpath, overwrite=overwrite, zarr_version = zarr_format)

        version = '0.5' if zarr_format == 3 else '0.4'

        meta = _get_or_create_multimeta(gr,
                                        axis_order=flataxes,
                                        unit_list=flatunit,
                                        version=version
                                        )

        meta.add_dataset(path=arrpath,
                         scale=flatscale,
                         overwrite=True)
        meta.retag(os.path.basename(dirpath))

        if flatchannels == 'auto':
            if 'c' in flataxes:
                idx = flataxes.index('c')
                size = arr.shape[idx]
            else:
                size = 1
            meta.autocompute_omerometa(size, arr.dtype)
        elif flatchannels is None:
            pass
        else:
            dtype = kwargs.get('dtype', arr.dtype)
            for channel in flatchannels:
                meta.add_channel(color = channel['color'],
                                 label = channel['label'],
                                 dtype = dtype
                                 )
        meta.save_changes()

        if verbose:
            print(f"Writer function: {writer_func}")
            print(f"Rechunk method: {rechunk_method}")

        results[key] = writer_func(arr=arr,
                                   store_path=key,  # compressor = compressor, dtype = dtype,
                                   chunks=chunks,
                                   shards = shards,
                                   dimension_names = flataxes,
                                   overwrite=overwrite,
                                   **kwargs
                                   )

    if compute:
        try:
            dask.compute(list(results.values()), retries = 6)
        except Exception as e:
            # print(e)
            pass
    else:
        return results
    return results



# Standard library imports
import copy
import logging
import os
import tempfile
import warnings
from pathlib import Path
from typing import Tuple, Union

# Third-party imports
import dask
import dask.array as da
import numpy as np

# Local application imports
from eubi_bridge.base.data_manager import BatchManager
from eubi_bridge.base.readers import read_single_image_asarray
from eubi_bridge.base.writers import store_arrays
# from eubi_bridge.fileset_io import FileSet
from eubi_bridge.fileset_io import BatchFile
from eubi_bridge.ngff.multiscales import Pyramid
from eubi_bridge.ngff.defaults import unit_map, scale_map, default_axes
from eubi_bridge.utils.convenience import (
    take_filepaths
)

# Configure logging
logging.getLogger('distributed.diskutils').setLevel(logging.CRITICAL)
warnings.filterwarnings('ignore')


class BridgeBase:
    def __init__(self,
                 input_path: Union[str, Path],  # TODO: add csv option (or a general table option).
                 includes=None,
                 excludes=None,
                 metadata_path=None,
                 series=None,
                 client=None,
                 zarr_format=2,
                 verbose=False
                 ):
        """
        Initialize the BridgeBase class. This class is the main entry point for
        converting and processing image data.

        Args:
            input_path (Union[str, Path]): Path to the input file or directory.
            includes (optional): Patterns of filenames to include.
            excludes (optional): Patterns of filenames to exclude.
            metadata_path (optional): Path to metadata file if any.
            series (optional): Series index or name to process.
            client (optional): Dask client for parallel processing.
            zarr_format (int, optional): Zarr format version. Defaults to 2.
            verbose (bool, optional): Enable verbose logging. Defaults to False.
        """

        # Ensure the input path is absolute
        if not input_path.startswith(os.sep):
            input_path = os.path.abspath(input_path)

        # Initialize instance variables
        self._input_path = input_path
        self._includes = includes
        self._excludes = excludes
        self._metadata_path = metadata_path
        self._series = series
        self._dask_temp_dir = None
        self._zarr_format = zarr_format
        self._verbose = verbose
        self.vmeta = None
        self._cluster_params = None
        self.client = client
        self.fileset = None
        self.pixel_metadata = None
        self.fileset = None

        # Validate the series parameter
        if self._series is not None:
            assert isinstance(self._series, (int, str)), (
                "The series parameter must be either an integer or string. "
                "Selection of multiple series from the same image is currently not supported."
            )

    def set_dask_temp_dir(self,
                          temp_dir: Union[str, Path] = 'auto'
                          ):
        """
        Set the temporary directory to store dask intermediate results.

        If the argument is 'auto' or None, the function will create a temporary directory
        in the system's default temporary directory and store the path in
        self._dask_temp_dir. If the argument is a string or Path, the function will create
        a temporary directory at the given path and store the path in
        self._dask_temp_dir.

        Parameters
        ----------
        temp_dir : Union[str, Path], optional
            The name of the temporary directory to store the intermediate results.
            Defaults to 'auto'.

        """
        if isinstance(temp_dir, tempfile.TemporaryDirectory):
            self._dask_temp_dir = temp_dir
            return
        if temp_dir in ('auto', None):
            temp_dir = tempfile.TemporaryDirectory(delete=False)
        elif isinstance(temp_dir, (str, Path)):
            os.makedirs(temp_dir, exist_ok=True)
            temp_dir = tempfile.TemporaryDirectory(dir=temp_dir, delete=False)
        else:
            raise TypeError(f"Invalid temp_dir argument: {temp_dir}")
        self._dask_temp_dir = temp_dir

    def read_dataset(self,
                     verified_for_cluster,
                     chunks_yx = None,
                     readers_params = {}
                     ):
        """
        - If the input path is a directory, can read single or multiple files from it.
        - If the input path is a file, can read a single image from it.
        - If the input path is a file with multiple series, can currently only read one series from it. Reading multiple series is currently not supported.
        :return:
        """
        input_path = self._input_path # todo: make them settable from this method?
        includes = self._includes
        excludes = self._excludes
        metadata_path = self._metadata_path
        series = self._series
        zarr_format = self._zarr_format
        verbose = self._verbose

        if os.path.isfile(input_path) or input_path.endswith('.zarr'):
            dirname = os.path.dirname(input_path)
            basename = os.path.basename(input_path)
            input_path = f"{dirname}/*{basename}"
            self._input_path = input_path

        self.filepaths = take_filepaths(input_path, includes, excludes)

        if series is None or series==0: # TODO: parallelize with cluster setting. Keep serial for no-cluster
            try:
                readers_params.pop('scene_index')
            except:
                pass

        futures = [read_single_image_asarray(path,
                                              chunks_yx=chunks_yx,
                                              verified_for_cluster=verified_for_cluster,
                                              zarr_format = zarr_format,
                                              verbose = verbose,
                                              **readers_params)
                   for path in self.filepaths]
        self.arrays = dask.compute(*futures)

        if metadata_path is None:
            self.metadata_path = self.filepaths[0]
        else:
            self.metadata_path = metadata_path

    def digest(self, # TODO: refactor to "assimilate_tags" and "concatenate?"
               time_tag: Union[str, tuple] = None,
               channel_tag: Union[str, tuple] = None,
               z_tag: Union[str, tuple] = None,
               y_tag: Union[str, tuple] = None,
               x_tag: Union[str, tuple] = None,
               axes_of_concatenation: Union[int, tuple, str] = None,
               ###
               metadata_reader: str = 'bfio',
               **kwargs
               ):
        """
        Digest the input data. This optionally involves
        concatenating multiple images along specified axes.

        Parameters
        ----------
        time_tag : Union[str, tuple], optional
            The tag for the time axis. Defaults to None.
        channel_tag : Union[str, tuple], optional
            The tag for the channel axis. Defaults to None.
        z_tag : Union[str, tuple], optional
            The tag for the z axis. Defaults to None.
        y_tag : Union[str, tuple], optional
            The tag for the y axis. Defaults to None.
        x_tag : Union[str, tuple], optional
            The tag for the x axis. Defaults to None.
        axes_of_concatenation : Union[int, tuple, str], optional
            The axes to concatenate. Defaults to None.


        Examples
        --------
            >>> # For a set of files with patterns in the file path in the following format:
            >>> # ["timepoint0_channel0_slice0.tif",...,"timepoint12_channel2_slice25.tif"] ]
            >>> # Concatenate only along specific axes:
            >>> bridge.digest(
            ...     time_tag='timepoint',
            ...     channel_tag='channel',
            ...     z_tag='slice_',
            ...     axes_of_concatenation='tz'  # Concatenate only time and z dimensions
            ... )
        """
        axes = 'tczyx'
        tags = (time_tag, channel_tag, z_tag, y_tag, x_tag)

        # Create a FileSet object
        self.batchfile = BatchFile(self.filepaths,
                               arrays=self.arrays,
                               axis_tag0=time_tag,
                               axis_tag1=channel_tag,
                               axis_tag2=z_tag,
                               axis_tag3=y_tag,
                               axis_tag4=x_tag,
                               )
        managers = self.batchfile.read_metadata(series=self._series,
                                                metadata_reader=metadata_reader,
                                                **kwargs)

        axdict = dict(zip(axes, tags))

        if axes_of_concatenation is None:
            axes_of_concatenation = []

        axlist = [axes.index(x) for x in axes_of_concatenation if x in axes]

        self.batchfile.batch_concatenate(axlist)
        # Get the refined arrays and sample paths
        (self.digested_arrays,
         self.digested_arrays_sample_paths,
         self.managers
         ) = self.batchfile.get_output_dicts(self._input_path)
        self._compute_pixel_metadata(**kwargs)
        return self


    def _compute_pixel_metadata(self,
                               **kwargs
                               ):
        """Compute and update pixel metadata for the digested arrays.

        Args:
            series: Series identifier
            metadata_reader: Reader to use for metadata (default: 'bfio')
            **kwargs: Additional metadata including units and scales
        """
        assert self.digested_arrays is not None
        assert self.digested_arrays_sample_paths is not None
        assert self.managers is not None

        unit_mapping = {
            'time_unit': 't', 'channel_unit': 'c',
            'z_unit': 'z', 'y_unit': 'y', 'x_unit': 'x'
        }
        scale_mapping = {
            'time_scale': 't', 'channel_scale': 'c',
            'z_scale': 'z', 'y_scale': 'y', 'x_scale': 'x'
        }

        # Process unit and scale updates
        update_unitdict = {unit_mapping[k]: v for k, v in kwargs.items() if k in unit_mapping}
        update_scaledict = {scale_mapping[k]: v for k, v in kwargs.items() if k in scale_mapping}

        # Update arrays and metadata
        for name, arr in self.digested_arrays.items():
            path = self.digested_arrays_sample_paths[name]
            manager = self.managers[name]
            manager.set_arraydata(arr)
            manager.update_meta(
                new_unitdict=update_unitdict,
                new_scaledict=update_scaledict
            )

        self.batchdata = BatchManager(self.managers)
        self.batchdata.fill_default_meta()

    def squeeze_dataset(self):
        self.batchdata.squeeze()

    def transpose_dataset(self,
                          dimension_order=Union[str, tuple, list]
                          ):
        """
        Transpose the dataset according to the given dimension order.

        Parameters
        ----------
        dimension_order : Union[str, tuple, list]
            The order of the dimensions in the transposed array.

        """
        self.batchdata.transpose(newaxes = dimension_order)

    def crop_dataset(self, **kwargs):
        self.batchdata.crop(**kwargs)

    def to_cupy(self):
        self.batchdata.to_cupy()


    def _prepare_array_metadata(self, batch_manager, sample_path_mapping):
        """Prepare metadata dictionaries for array storage.

        Args:
            batch_manager: BatchManager instance containing array data
            sample_path_mapping: Dictionary mapping array names to their file paths

        Returns:
            Tuple containing dictionaries for arrays, scales, axes, units, and chunks
        """
        array_data = {}
        dimension_scales = {}
        dimension_axes = {}
        dimension_units = {}
        chunk_configs = {}
        channel_meta = {} # a dict of lists, each list being the length of channels per image

        for array_name, file_path in sample_path_mapping.items():
            manager = batch_manager.managers[array_name]
            array_data[array_name] = {'0': manager.array}
            dimension_scales[array_name] = {'0': manager.scales}
            dimension_axes[array_name] = {'0': manager.axes}
            dimension_units[array_name] = {'0': manager.units}
            chunk_configs[array_name] = {'0': manager.chunks}
            channel_meta[array_name] = {'0': manager.channels}

        return array_data, dimension_scales, dimension_axes, dimension_units, chunk_configs, channel_meta


    def _create_output_path_mapping(self, output_dir,
                                    nested_data, sample_paths):
        """Create flattened path mappings for the given data dictionary.

        Args:
            output_dir: Base output directory
            nested_data: Nested dictionary of data to be flattened
            sample_paths: Dictionary mapping array names to their file paths

        Returns:
            Dictionary with output file paths as keys
        """
        return {
            os.path.join(
                output_dir,
                f"{array_name}.zarr" if not array_name.endswith('zarr') else array_name,
                str(level)
            ): value
            for array_name, subdict in nested_data.items()
            for level, value in subdict.items()
        }


    def _process_chunking_configurations(self,
                                         chunk_sizes,
                                         shard_coefficients,
                                         axis_mappings,
                                         chunk_mappings):
        """Process chunk and shard configurations for each array.

        Args:
            chunk_sizes: Dictionary of chunk sizes per dimension
            shard_coefficients: Dictionary of shard coefficients per dimension
            axis_mappings: Dictionary mapping output paths to their dimension axes
            chunk_mappings: Dictionary mapping output paths to their chunk configurations

        Returns:
            Tuple of (updated_chunk_sizes, updated_shard_coefficients)
        """
        processed_chunk_sizes = {}
        processed_shard_coeffs = {}

        for output_path, chunk_config in chunk_mappings.items():
            axes = axis_mappings[output_path]
            final_chunk_sizes = []
            final_shard_coeffs = []

            for axis in axes:
                chunk_size = chunk_sizes[axis] or chunk_config[axes.index(axis)]
                final_chunk_sizes.append(chunk_size)
                final_shard_coeffs.append(shard_coefficients[axis])

            processed_chunk_sizes[output_path] = final_chunk_sizes
            processed_shard_coeffs[output_path] = final_shard_coeffs

        return processed_chunk_sizes, processed_shard_coeffs


    def write_arrays(self,
                     output_dir,
                     compute=True,
                     use_tensorstore=False,
                     rechunk_method='tasks',
                     **kwargs):
        """Write processed arrays to storage.

        Args:
            output_dir: Base output directory
            compute: Whether to compute the result immediately
            use_tensorstore: Whether to use tensorstore for storage
            rechunk_method: Method to use for rechunking ('auto', 'rechunker', etc.)
            **kwargs: Additional arguments for array storage

        Returns:
            Results of the storage operation
        """
        output_dir = os.path.abspath(output_dir)
        storage_options = kwargs.copy()

        # Apply transformations
        if storage_options.get('use_gpu', False):
            self.to_cupy()
        if storage_options.get('squeeze', False):
            self.squeeze_dataset()
        if storage_options.get('dimension_order'):
            self.transpose_dataset(storage_options['dimension_order'])

        # Update storage options with format and verbosity
        storage_options.update({
            'zarr_format': self._zarr_format,
            'verbose': self._verbose
        })

        # Apply cropping (will not do anything if no cropping parameter is provided)
        self.crop_dataset(**storage_options)

        # Prepare data for storage
        batch_manager = self.batchdata
        sample_path_mapping = self.digested_arrays_sample_paths
        assert batch_manager is not None, "The 'batchdata' must be computed before writing arrays"

        # Extract array metadata
        (array_data,
         dimension_scales,
         dimension_axes,
         dimension_units,
         chunk_configs,
         channel_meta) = self._prepare_array_metadata(
            batch_manager, sample_path_mapping
        )

        # Create path mappings for storage
        path_mappings = {
            'arrays': self._create_output_path_mapping(output_dir, array_data, sample_path_mapping),
            'scales': self._create_output_path_mapping(output_dir, dimension_scales, sample_path_mapping),
            'axes': self._create_output_path_mapping(output_dir, dimension_axes, sample_path_mapping),
            'units': self._create_output_path_mapping(output_dir, dimension_units, sample_path_mapping),
            'chunks': self._create_output_path_mapping(output_dir, chunk_configs, sample_path_mapping),
            'channels':self._create_output_path_mapping(output_dir, channel_meta, sample_path_mapping),
        }

        # Configure chunking and sharding
        chunk_sizes = {
            't': storage_options.get('time_chunk'),
            'c': storage_options.get('channel_chunk'),
            'z': storage_options.get('z_chunk'),
            'y': storage_options.get('y_chunk'),
            'x': storage_options.get('x_chunk')
        }

        shard_coefficients = {
            't': storage_options.get('time_shard_coef', 1),
            'c': storage_options.get('channel_shard_coef', 1),
            'z': storage_options.get('z_shard_coef', 3),
            'y': storage_options.get('y_shard_coef', 3),
            'x': storage_options.get('x_shard_coef', 3)
        }

        # Process chunk configurations
        processed_chunk_sizes, processed_shard_coeffs = self._process_chunking_configurations(
            chunk_sizes,
            shard_coefficients,
            path_mappings['axes'],
            path_mappings['chunks']
        )

        # Store arrays
        storage_results = store_arrays(
            path_mappings['arrays'],
            output_dir,
            axes=path_mappings['axes'],
            scales=path_mappings['scales'],
            units=path_mappings['units'],
            output_chunks=processed_chunk_sizes,
            output_shard_coefficients=processed_shard_coeffs,
            use_tensorstore=use_tensorstore,
            compute=compute,
            rechunk_method=rechunk_method,
            channel_meta = path_mappings['channels'] or None,
            **storage_options
        )

        self.flatarrays = path_mappings['arrays']

        # Save OME-XML metadata if requested
        if storage_options.get('save_omexml', False):
            manager_paths = {
                os.path.join(output_dir, f"{name}.zarr" if not name.endswith('zarr') else name):
                    batch_manager.managers[name]
                for name, file_path in sample_path_mapping.items()
            }

            for output_path, manager in manager_paths.items():
                if manager.omemeta is None:
                    manager.create_omemeta()
                manager.save_omexml(output_path)

        return storage_results

def downscale(
        gr_paths,
        time_scale_factor,
        channel_scale_factor,
        z_scale_factor,
        y_scale_factor,
        x_scale_factor,
        n_layers,
        downscale_method='simple',
        **kwargs
        ):

    scale_factor_dict = {
                        't': time_scale_factor,
                        'c': channel_scale_factor,
                        'z': z_scale_factor,
                        'y': y_scale_factor,
                        'x': x_scale_factor
                         }

    if isinstance(gr_paths, dict):
        gr_paths = list(set(os.path.dirname(key) for key in gr_paths.keys()))

    pyrs = [Pyramid(path) for path in gr_paths] # TODO: add a to_cupy parameter here.
    result_collection = []

    for pyr in pyrs:
        scale_factor = [scale_factor_dict[ax] for ax in pyr.meta.axis_order]

        pyr.update_downscaler(scale_factor=scale_factor,
                              n_layers=n_layers,
                              downscale_method=downscale_method
                              )
        grpath = pyr.gr.store.root
        grname = os.path.basename(grpath)
        grdict = {grname: {}}
        axisdict = {grname: {}}
        scaledict = {grname: {}}
        unitdict = {grname: {}}
        chunkdict = {grname: {}}
        sharddict = {grname: {}}

        for key, value in pyr.downscaler.downscaled_arrays.items():
            if key != '0':
                grdict[grname][key] = value
                axisdict[grname][key] = tuple(pyr.meta.axis_order)
                scaledict[grname][key] = tuple(pyr.downscaler.dm.scales[int(key)])
                unitdict[grname][key] = tuple(pyr.meta.unit_list)
                chunkdict[grname][key] = tuple(pyr.base_array.chunksize)
                # channeldict[grname][key] = tuple(pyr.meta.channels)
                if pyr.meta.zarr_format == 3:
                    sharddict[grname][key] = tuple(pyr.meta.zarr_group)
                    basepath = pyr.meta.resolution_paths[0]
                    sharddict[grname][key] = tuple(pyr.layers[basepath].shards)

        output_path = os.path.dirname(grpath)
        arrays = {k: {'0': v} if not isinstance(v, dict) else v for k, v in grdict.items()}

        ### TODO: make this a separate function? def flatten_pyramids
        flatarrays = {os.path.join(output_path, f"{key}.zarr"
                      if not key.endswith('zarr') else key, str(level)): arr
                      for key, subarrays in arrays.items()
                      for level, arr in subarrays.items()}
        flataxes = {os.path.join(output_path, f"{key}.zarr"
                      if not key.endswith('zarr') else key, str(level)): axes
                      for key, subaxes in axisdict.items()
                      for level, axes in subaxes.items()}
        flatscales = {os.path.join(output_path, f"{key}.zarr"
                      if not key.endswith('zarr') else key, str(level)): scale
                      for key, subscales in scaledict.items()
                      for level, scale in subscales.items()}
        flatunits = {os.path.join(output_path, f"{key}.zarr"
                      if not key.endswith('zarr') else key, str(level)): unit
                      for key, subunits in unitdict.items()
                      for level, unit in subunits.items()}
        flatchunks = {os.path.join(output_path, f"{key}.zarr"
                      if not key.endswith('zarr') else key, str(level)): chunk
                      for key, subchunks in chunkdict.items()
                      for level, chunk in subchunks.items()}

        if len(sharddict) > 0:
            flatshards = {os.path.join(output_path, f"{key}.zarr"
                          if not key.endswith('zarr') else key, str(level)): shard
                          for key, subshards in sharddict.items()
                          for level, shard in subshards.items()}
        else:
            flatshards = None
        ### TODO ends

        results = store_arrays(flatarrays,
                               output_path=output_path,
                               axes = flataxes,
                               scales=flatscales,
                               units=flatunits,
                               output_chunks = flatchunks,
                               output_shards = flatshards,
                               compute=False,
                               channel_meta=None,
                               **kwargs
                               )

        result_collection += list(results.values())
    if 'rechunk_method' in kwargs:
        if kwargs.get('rechunk_method') == 'rechunker':
            raise NotImplementedError(f"Rechunker is not supported for the downscaling step.")
    if 'max_mem' in kwargs:
        raise NotImplementedError(f"Rechunker is not supported for the downscaling step.")
    try:
        dask.compute(*result_collection)
    except Exception as e:
        # print(e)
        pass
    return results



import zarr
from pathlib import Path
from typing import Optional, Dict, List, Any, Union, Iterable, ClassVar
from dataclasses import dataclass
import copy
import numpy as np
import dask.array as da
from eubi_bridge.base.scale import Downscaler


def is_zarr_group(path: (str, Path)
                  ):
    try:
        _ = zarr.open_group(path, mode = 'r')
        return True
    except:
        return False


def generate_channel_metadata(num_channels,
                              dtype = np.uint16
                              ):
    # Standard distinct microscopy colors
    default_colors = [
        "FF0000",  # Red
        "00FF00",  # Green
        "0000FF",  # Blue
        "FF00FF",  # Magenta
        "00FFFF",  # Cyan
        "FFFF00",  # Yellow
        "FFFFFF",  # White
    ]

    channels = []
    import numpy as np
    min = 0
    max = np.iinfo(dtype).max
    for i in range(num_channels):
        color = default_colors[i] if i < len(default_colors) else f"{i*40%256:02X}{i*85%256:02X}{i*130%256:02X}"
        channel = {
            "color": color,
            "coefficient": 1,
            "active": True,
            "label": f"Channel {i}",
            "window": {
                "min": min,
                "max": max,
                "start": min,
                "end": max
            },
            "family": "linear",
            "inverted": False
        }
        channels.append(channel)

    return {
        "omero": {
            "channels": channels,
            "rdefs": {
                "defaultT": 0,
                "model": "greyscale",
                "defaultZ": 0
            }
        }
    }


class NGFFMetadataHandler:
    """Class for handling NGFF metadata in zarr groups."""

    SUPPORTED_VERSIONS: ClassVar[List[str]] = ["0.4", "0.5"]

    def __init__(self) -> None:
        """Initialize an empty metadata handler."""
        self.zarr_group: Optional[zarr.Group] = None
        self.metadata: Optional[Dict[str, Any]] = None
        self._pending_changes: bool = False
        self.version: Optional[str] = None
        self.zarr_format: Optional[int] = None

    def __enter__(self) -> 'NGFFMetadataHandler':
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        if self._pending_changes:
            self.save_changes()

    @property
    def multiscales(self) -> Dict[str, Any]:
        """Get the multiscales metadata."""
        if not self.metadata or 'multiscales' not in self.metadata:
            raise RuntimeError("No multiscales metadata available")
        return self.metadata['multiscales'][0]

    @property
    def omero(self) -> Dict[str, Any]:
        """Get the multiscales metadata."""
        if not self.metadata or 'omero' not in self.metadata:
            raise RuntimeError("No omero metadata available")
        return self.metadata['omero']

    def _validate_version_and_format(self, version: str, zarr_format: int) -> None:
        """Validate version and zarr format compatibility."""
        if version not in self.SUPPORTED_VERSIONS:
            raise ValueError(f"Unsupported version {version}. Supported versions: {self.SUPPORTED_VERSIONS}")
        if zarr_format not in (2, 3):
            raise ValueError(f"Unsupported Zarr format: {zarr_format}")
        if version == "0.5" and zarr_format != 3:
            raise ValueError("NGFF version 0.5 requires Zarr format 3")

    def _validate_axis_inputs(self, axis_order: str, units: Optional[List[str]]) -> None:
        """Validate axis order and units inputs."""
        if not all(ax in 'tczyx' for ax in axis_order):
            raise ValueError("Invalid axis order. Must contain only t,c,z,y,x")
        if units is not None:
            if not (len(axis_order) - len(units)) in [0, 1]:
                raise ValueError("Number of units must match number of axes except channel")
            elif (len(axis_order) - len(units)) == 1:
                if 'c' not in axis_order:
                    raise ValueError("Only channel axis can be kept without a unit.")

    def _get_dataset(self, path: str) -> Optional[Dict[str, Any]]:
        """Helper method to find dataset by path."""
        path = str(path)
        for dataset in self.multiscales['datasets']:
            if dataset['path'] == path:
                return dataset
        return None

    def _update_coordinate_transformation(self,
                                          dataset: Dict[str, Any],
                                          transform_type: str,
                                          values: List[float]) -> None:
        """Update or add a coordinate transformation."""
        for transform in dataset['coordinateTransformations']:
            if transform['type'] == transform_type:
                transform[transform_type] = values
                break
        else:
            if transform_type == 'scale':
                dataset['coordinateTransformations'].insert(
                    0, {'type': transform_type, transform_type: values}
                )
            else:
                dataset['coordinateTransformations'].append(
                    {'type': transform_type, transform_type: values}
                )

    def get_metadata_state(self) -> Dict[str, Any]:
        """Get a copy of current metadata state."""
        if self.metadata is None:
            raise RuntimeError("No metadata loaded or created")
        return copy.deepcopy(self.metadata)

    def get_summary(self) -> Dict[str, Any]:
        """Get a summary of the metadata."""
        if not self.metadata:
            raise RuntimeError("No metadata available")

        return {
            'version': self.version,
            'zarr_format': self.zarr_format,
            'axes': self._axis_names,
            'units': self._units,
            'n_datasets': len(self.multiscales['datasets']),
            'name': self.multiscales['name']
        }


    def create_new(self, version: str = "0.5", name: str = "Series 0") -> 'NGFFMetadataHandler':
        """Create a new metadata handler with empty metadata of specified version."""
        self._validate_version_and_format(version, 3 if version == "0.5" else 2)

        multiscale_metadata = {
            'name': name,
            'axes': [],
            'datasets': [],
            'metadata': {}
        }

        if version == "0.5":
            self.metadata = {
                'version': version,
                'multiscales': [multiscale_metadata],
                'omero': {
                    'channels': [],
                    'rdefs': {
                        'defaultT': 0,
                        'model': 'greyscale',
                        'defaultZ': 0
                    }
                },
                '_creator': {
                    'name': 'NGFFMetadataHandler',
                    'version': '1.0'
                }
            }
        else: # version == "0.4"
            multiscale_metadata['version'] = version
            self.metadata = {
                '_creator': {
                    'name': 'NGFFMetadataHandler',
                    'version': '1.0'
                },
                'multiscales': [multiscale_metadata],
                'omero': {
                    'channels': [],
                    'rdefs': {
                        'defaultT': 0,
                        'model': 'greyscale',
                        'defaultZ': 0
                    }
                }
            }

        self.version = version
        self.zarr_format = 3 if version == "0.5" else 2
        self._pending_changes = True
        return self

    def connect_to_group(self, store: Union[zarr.Group, str, Path], mode: str = 'a') -> None:
        """Connect to a zarr group for reading/writing metadata."""
        if not isinstance(store, (zarr.Group, str, Path)):
            raise ValueError("Store must be a zarr group or path")
        if isinstance(store, zarr.Group):
            self.zarr_group = store
        else: # isinstance(store, (str, Path))
            if is_zarr_group(store):
                self.zarr_group = zarr.open_group(store, mode=mode)
                # zarr_version = self.zarr_group.info._zarr_format
            else:
                zarr_version = self.zarr_format if self.zarr_format else 2
                self.zarr_group = zarr.open_group(store, mode=mode, zarr_version=zarr_version)
        # Update handler's format to match the created store
        store_format = self.zarr_group.info._zarr_format
        self.zarr_format = store_format
        # Update version based on zarr_format
        self.version = "0.5" if store_format == 3 else "0.4"
        self._validate_version_and_format(self.version, store_format)

    def read_metadata(self):
        """Read metadata from connected zarr group."""
        if self.zarr_group is None:
            raise RuntimeError("No zarr group connected. Call connect_to_group first.")

        if 'ome' in self.zarr_group.attrs:
            self.metadata = self.zarr_group.attrs['ome']
            self.version = self.metadata['version']
        elif 'multiscales' in self.zarr_group.attrs:
            self.metadata = {'multiscales': self.zarr_group.attrs['multiscales']}
            self.version = self.metadata['multiscales'][0]['version']
        else:
            raise ValueError("No valid metadata found in zarr group")

        if 'omero' in self.zarr_group.attrs:
            self.metadata['omero'] = self.zarr_group.attrs['omero']
        self.zarr_format = 3 if self.version == "0.5" else 2
        self._pending_changes = False
        return self

    def save_changes(self) -> None:
        """Save current metadata to connected zarr group."""
        if not self._pending_changes:
            return
        if self.zarr_group is None:
            raise RuntimeError("No zarr group connected. Call connect_to_group first.")

        if self.metadata.get('version', '') == '0.5':
            self.zarr_group.attrs['ome'] = self.metadata
        else:
            self.zarr_group.attrs['multiscales'] = self.metadata['multiscales']
            if 'omero' in self.metadata:
                # print(f"omero meta: {self.metadata['omero']}")
                self.zarr_group.attrs['omero'] = self.metadata['omero']
            if '_creator' in self.metadata:
                self.zarr_group.attrs['_creator'] = self.metadata['_creator']

        self._pending_changes = False

    def update_all_datasets(self,
                            scale: Optional[List[float]] = None,
                            translation: Optional[List[float]] = None
                            ) -> None:
        """Update all datasets with new scale and/or translation values."""
        for dataset in self.multiscales['datasets']:
            if scale is not None:
                self._update_coordinate_transformation(dataset, 'scale', scale)
            if translation is not None:
                self._update_coordinate_transformation(dataset, 'translation', translation)
        self._pending_changes = True

    def autocompute_omerometa(self,
                              n_channels: int,
                              dtype
                              ) -> None:
        """Add multiple channels to the OMERO metadata."""
        omero_meta = generate_channel_metadata(n_channels, dtype)
        self.metadata['omero'] = omero_meta['omero']
        self._pending_changes = True


    def add_channel(self,
                    color: str = "808080",
                    label: str = None,
                    dtype = None) -> None:
        """Add a channel to the OMERO metadata."""

        # import numpy as np
        assert dtype is not None
        min = 0
        max = np.iinfo(dtype).max
        
        if 'omero' not in self.metadata:
            self.metadata['omero'] = {
                'channels': [],
                'rdefs': {
                    'defaultT': 0,
                    'model': 'greyscale',
                    'defaultZ': 0
                }
            }

        channel = {
            'color': color,
            'coefficient': 1,
            'active': True,
            'label': label or f"Channel {len(self.metadata['omero']['channels'])}",
            'window': {'min': min, 'max': max, 'start': min, 'end': max},
            'family': 'linear',
            'inverted': False
        }

        self.metadata['omero']['channels'].append(channel)
        self._pending_changes = True

    def get_channels(self) -> List[Dict[str, Any]]:
        """
        Get a list of all channels with their labels and colors.

        Returns:
            List[Dict[str, Any]]: A list of dictionaries, where each dictionary contains
                                 'label' and 'color' keys for a channel.
        """
        if 'omero' not in self.metadata or 'channels' not in self.metadata['omero']:
            return []

        return [
            {
                'label': channel.get('label', f"Channel {i}"),
                'color': channel.get('color', '808080')
            }
            for i, channel in enumerate(self.metadata['omero']['channels'])
        ]

    def parse_axes(self, ###
                   axis_order: str,
                   units: Optional[List[str]] = None) -> None:
        """Update axes information with new axis order and units."""
        if self.metadata is None:
            raise RuntimeError("No metadata loaded or created.")

        self._validate_axis_inputs(axis_order, units)

        if units is None:
            units = [None] * len(axis_order)
        if len(axis_order) - len(units) == 1:
            if 'c' in axis_order:
                idx = axis_order.index('c')
                units.insert(idx, None)

        new_axes = []
        for ax_name, unit in zip(axis_order, units):
            axis_data = {
                'name': ax_name,
                'type': {'t': 'time', 'c': 'channel', 'z': 'space',
                         'y': 'space', 'x': 'space'}.get(ax_name, 'custom')
            }
            if unit is not None:
                axis_data['unit'] = unit
            new_axes.append(axis_data)

        self.metadata['multiscales'][0]['axes'] = new_axes
        self._pending_changes = True

    def add_dataset(self, path: Union[str, int],
                    scale: Iterable[Union[int, float]],
                    translation: Optional[Iterable[Union[int, float]]] = None,
                    overwrite: bool = False) -> None:
        """Add a dataset with scale and optional translation."""
        path = str(path)
        scale = list(map(float, scale))
        if translation is not None:
            translation = list(map(float, translation))

        dataset_data = {
            'path': path,
            'coordinateTransformations': [{'type': 'scale', 'scale': scale}]
        }
        if translation is not None:
            dataset_data['coordinateTransformations'].append(
                {'type': 'translation', 'translation': translation}
            )

        existing_paths = self.get_resolution_paths()
        if path in existing_paths:
            if not overwrite:
                raise ValueError(f"Dataset path '{path}' already exists")
            idx = existing_paths.index(path)
            self.metadata['multiscales'][0]['datasets'][idx] = dataset_data
        else:
            self.metadata['multiscales'][0]['datasets'].append(dataset_data)

        self.metadata['multiscales'][0]['datasets'].sort(
            key=lambda x: int(x['path']) if x['path'].isdigit() else float('inf')
        )
        self._pending_changes = True

    def update_scale(self,
                     path: Union[str, int],
                     scale: Iterable[Union[int, float]]) -> None:
        """Update scale for a specific dataset."""
        dataset = self._get_dataset(str(path))
        if dataset:
            self._update_coordinate_transformation(dataset, 'scale', list(map(float, scale)))
            self._pending_changes = True

    def update_translation(self, path: Union[str, int],
                           translation: Iterable[Union[int, float]]) -> None:
        """Update translation for a specific dataset."""
        dataset = self._get_dataset(str(path))
        if dataset:
            self._update_coordinate_transformation(dataset, 'translation', list(map(float, translation)))
            self._pending_changes = True

    def get_resolution_paths(self) -> List[str]:
        """Get paths to all resolution levels."""
        return [ds['path'] for ds in self.multiscales['datasets']]

    @property
    def _axis_names(self) -> List[str]:
        """Get list of axis names."""
        return [ax['name'] for ax in self.multiscales['axes']]

    @property
    def axis_order(self) -> str:
        """Get axis names as str."""
        return ''.join(self._axis_names)

    @property
    def _units(self) -> Dict[str, Optional[str]]:
        """Get dictionary of axis units."""
        return {ax['name']: ax.get('unit') for ax in self.multiscales['axes']}

    @property
    def unit_dict(self):
        return self._units

    @property
    def unit_list(self):
        return [self._units[ax] for ax in self._axis_names]

    @property
    def ndim(self) -> int:
        return len(self.axis_order)

    @property
    def resolution_paths(self) -> List[str]:
        return [item['path'] for item in self.multiscales['datasets']]

    @property
    def nlayers(self) -> int:
        return len(self.resolution_paths)

    @property
    def channels(self):
        return self.get_channels()


    def validate_metadata(self) -> bool:
        """Validate current metadata structure."""
        if not self.metadata:
            return False

        try:
            if self.version == "0.5":
                if not all(key in self.metadata for key in {'version', 'multiscales'}):
                    return False
            else:  # version == "0.4"
                if 'multiscales' not in self.metadata:
                    return False
                if 'version' not in self.metadata['multiscales'][0]:
                    return False

            required_keys = {'name', 'axes', 'datasets'}
            return all(key in self.multiscales for key in required_keys)

        except (KeyError, IndexError, TypeError):
            return False

    ###

    def get_scaledict(self,
                      pth: Union[str, int]
                      ):
        # pth = cnv.asstr(pth)
        idx = self.resolution_paths.index(pth)
        scale = self.multiscales['datasets'][idx]['coordinateTransformations'][0]['scale']
        return dict(zip(self.axis_order, scale))

    def get_base_scaledict(self):
        basepath = self.resolution_paths[0]
        return self.get_scaledict(basepath)

    def get_scale(self,
                  pth: Union[str, int]
                  ):
        scaledict = self.get_scaledict(pth)
        return [scaledict[ax] for ax in self.axis_order]

    def get_base_scale(self):
        basepath = self.resolution_paths[0]
        return self.get_scale(basepath)

    def set_scale(self,
                  pth: Union[str, int] = 'auto',
                  scale: Union[tuple, list, dict] = 'auto',
                  # hard=False
                  ):
        if isinstance(scale, tuple):
            scale = list(scale)
            ch_index = self.axis_order.index('c')
            scale[ch_index] = 1
        elif hasattr(scale, 'tolist'):
            scale = scale.tolist()
        elif isinstance(scale, dict):  # TODO: test this block further
            assert all([ax in self.axis_order for ax in scale])
            fullscale = self.get_scale(pth)
            scaledict = dict(zip(self.axis_order, fullscale))
            scaledict.update(**scale)
            scale = [scaledict[ax] for ax in self.axis_order]

        if pth == 'auto':
            pth = self.resolution_paths[0]
        if scale == 'auto':
            pth = self.scales[pth]
        idx = self.resolution_paths.index(pth)
        self.multiscales['datasets'][idx]['coordinateTransformations'][0]['scale'] = scale
        self._pending_changes = True
        # if hard:
        #     self.gr.attrs['multiscales'] = self.multimeta
        return

    def update_scales(self,
                      reference_scale: Union[tuple, list],  # , dict],
                      scale_factors: dict,
                      # hard=True
                      ):
        for pth, factor in scale_factors.items():
            new_scale = np.multiply(factor, reference_scale)
            self.set_scale(pth, new_scale)#,hard)
        return self

    def update_unitlist(self,
                        unitlist = None,
                        # hard=False
                        ):
        if isinstance(unitlist, tuple):
            unitlist = list(unitlist)
        assert isinstance(unitlist, list)
        self.parse_axes(self.axis_order, unitlist#overwrite=True
        )
        # if hard:
        #     self.gr.attrs['multiscales'] = self.multimeta
        return self

    @property
    def scales(self):
        scales = {}
        for pth in self.resolution_paths:
            scl = self.get_scale(pth)
            scales[pth] = scl
        return scales

    @property
    def scaledict(self):
        scales = {}
        for pth in self.resolution_paths:
            scl = self.get_scaledict(pth)
            scales[pth] = scl
        return scales

    def retag(self,
              new_tag: str,
              ):
        self.multiscales['name'] = new_tag
        self._pending_changes = True
        return self




class Pyramid:
    def __init__(self,
                 gr: (zarr.Group, zarr.storage.StoreLike, Path, str) = None
                 # An NGFF group. This contains the multiscales metadata in attrs and image layers as
                 ):
        self.meta = None
        self.gr = None
        if gr is not None:
            self.from_ngff(gr)

    def __repr__(self):
        return f"NGFF with {self.nlayers} layers."

    def from_ngff(self, gr):
        self.meta = NGFFMetadataHandler()
        self.meta.connect_to_group(gr)
        self.meta.read_metadata()
        self.gr = self.meta.zarr_group
        return self

    def to_ngff(self,
                store: (zarr.Group, zarr.storage.StoreLike, Path, str),
                version: str = "0.5"
                ):
        newmeta = NGFFMetadataHandler()
        if is_zarr_group(store):
            self.gr = zarr.open_group(store, mode='a')
            newmeta.connect_to_group(self.gr)
        else:
            self.meta.create_new(version = version, name = "Series 0")
        newmeta.save_changes()
        self.meta = newmeta
        return self

    @property
    def axes(self):
        return self.meta.axis_order

    @property
    def nlayers(self):
        return self.meta.nlayers

    @property
    def layers(self):
        return {path: self.gr[path] for path in self.gr.array_keys()}

    def get_dask_data(self):
        return {str(path): da.from_zarr(self.layers[path]) for path in self.gr.array_keys()}

    @property
    def dask_arrays(self):
        return self.get_dask_data()

    @property
    def base_array(self):
        return self.dask_arrays['0']

    # def shrink(self,
    #            hard=False
    #            ):
    #     """
    #     Delete all arrays except for the base array (zeroth array)
    #     :return:
    #     """
    #     for key in self.meta.resolution_paths:
    #         if key == '0':
    #             continue
    #         self.meta.del_dataset(key)
    #         if hard:
    #             del self.gr[key]
    #     if hard:
    #         self.meta.to_ngff(self.gr)
    #     return

    def update_scales(self,
                      **kwargs
                      ):
        """
        Automatically updates all pixel values for all layers based on
        provided pixel values for specific axes corresponding to
        the top resolution layer.
        :param kwargs:
        :return:
        """
        hard = kwargs.get('hard', False)

        new_scaledict = self.meta.get_base_scaledict()
        for ax in self.meta.axis_order:
            if ax in kwargs:
                new_scaledict[ax] = kwargs.get(ax)
        new_scale = [new_scaledict[ax] for ax in self.meta.axis_order]
        ###
        shapes = [self.layers[key].shape for key in self.meta.resolution_paths]
        scale_factors = np.divide(shapes[0], shapes)
        scale_factordict = {pth: scale
                            for pth, scale in
                            zip(self.meta.resolution_paths, scale_factors.tolist())}
        self.meta.update_scales(reference_scale=new_scale,
                                scale_factors=scale_factordict,
                                )
        if hard:
            self.meta.save_changes()
        return

    def update_units(self,
                     **kwargs
                     ):
        """
        Automatically updates all pixel units based on provided unit strings for each axis.
        :param kwargs:
        :return:
        """
        hard = kwargs.get('hard', False)

        new_unitdict = self.meta.unit_dict
        for ax in self.meta.axis_order:
            if ax in kwargs:
                new_unitdict[ax] = kwargs.get(ax)
        new_unitlist = [new_unitdict[ax] for ax in self.meta.axis_order]
        ###
        self.meta.update_unitlist(unitlist=new_unitlist,
                                  # hard=hard
                                  )

    @property
    def tag(self):
        return self.multimeta[0]['name']

    def retag(self,
              new_tag: str,
              hard = False
              ):
        self.meta.retag(new_tag)
        if hard:
            self.meta.save_changes()
        return self

    def update_downscaler(self,
                          scale_factor=None,
                          n_layers=1,
                          downscale_method='simple',
                          backend='numpy',
                          **kwargs
                          ):
        darr = self.base_array
        if scale_factor is None:
            scale_factor = tuple([defaults.scale_factor_map[key] for key in self.axes])
        scale = self.meta.scales['0']
        scale_factor = tuple(np.minimum(darr.shape, scale_factor))
        self.downscaler = Downscaler(array=darr,
                                     scale_factor=scale_factor,
                                     n_layers=n_layers,
                                     scale=scale,
                                     downscale_method=downscale_method,
                                     backend=backend
                                     )
        return self


import shutil, ctypes, time, os, zarr, pprint, psutil, dask, copy
import numpy as np, os, glob, tempfile
from multiprocessing.pool import ThreadPool

from dask import array as da
from distributed import LocalCluster, Client
from dask_jobqueue import SLURMCluster
from pathlib import Path
from typing import Union

# from eubi_bridge.ngff.multiscales import Pyramid
# from eubi_bridge.ngff import defaults
from eubi_bridge.base.data_manager import BatchManager
from eubi_bridge.ebridge_base import BridgeBase, downscale
from eubi_bridge.utils.convenience import take_filepaths, is_zarr_group
from eubi_bridge.utils.metadata_utils import print_printable, get_printables
from eubi_bridge.utils.logging_config import get_logger

import logging, warnings

# Set up logger for this module
logger = get_logger(__name__)

# Suppress noisy logs
logging.getLogger('distributed.diskutils').setLevel(logging.CRITICAL)
logging.getLogger('distributed.worker').setLevel(logging.WARNING)
logging.getLogger('distributed.scheduler').setLevel(logging.WARNING)

def verify_filepaths_for_cluster(filepaths):
    """Verify that all file extensions are supported for distributed processing."""
    logger.info("Verifying file extensions for distributed setup.")
    formats = ['lif', 'czi', 'lsm',
               'nd2', 'ome.tiff', 'ome.tif',
               'tiff', 'tif', 'zarr',
               'png', 'jpg', 'jpeg']
    
    for filepath in filepaths:
        verified = any(list(map(lambda path, ext: path.endswith(ext), [filepath] * len(formats), formats)))
        if not verified:
            root, ext = os.path.splitext(filepath)
            logging.warning(f"Distributed execution is not supported for the {ext} format")
            logger.warning(f"Falling back on multithreading.")
            break
    if verified:
        logger.info("File extensions were verified for distributed setup.")
    return verified

class EuBIBridge:
    """
    EuBIBridge is a conversion tool for bioimage datasets, allowing for both unary and aggregative conversion of image
    data collections to OME-Zarr format.

    Attributes:
        config_gr (zarr.Group): Configuration settings stored in a Zarr group.
        config (dict): Dictionary representation of configuration settings for cluster, conversion, and downscaling.
        dask_config (dict): Dictionary representation of configuration settings for dask.distributed.
        root_defaults (dict): Installation defaults of configuration settings for cluster, conversion, and downscaling.
        root_dask_defaults (dict): Installation defaults of configuration settings for dask.distributed.
    """
    def __init__(self,
                 configpath = f"{os.path.expanduser('~')}/.eubi_bridge",
               ):
        """
        Initializes the EuBIBridge class and loads or sets up default configuration.

        Args:
            configpath (str, optional): Path to store configuration settings. Defaults to the home directory.
        """

        root_dask_defaults = {'distributed.adaptive.interval': '1s', 'distributed.adaptive.maximum': '.inf',
             'distributed.adaptive.minimum': 0, 'distributed.adaptive.target-duration': '5s',
             'distributed.adaptive.wait-count': 3, 'distributed.admin.event-loop': 'tornado',
             'distributed.admin.large-graph-warning-threshold': '10MB',
             'distributed.admin.log-format': '%(asctime)s - %(name)s - %(levelname)s - %(message)s',
             'distributed.admin.log-length': 10000, 'distributed.admin.low-level-log-length': 1000,
             'distributed.admin.max-error-length': 10000, 'distributed.admin.pdb-on-err': False,
             'distributed.admin.system-monitor.disk': True, 'distributed.admin.system-monitor.gil.enabled': True,
             'distributed.admin.system-monitor.gil.interval': '1ms', 'distributed.admin.system-monitor.host-cpu': False,
             'distributed.admin.system-monitor.interval': '500ms', 'distributed.admin.system-monitor.log-length': 7200,
             'distributed.admin.tick.cycle': '1s', 'distributed.admin.tick.interval': '20ms',
             'distributed.admin.tick.limit': '3s', 'distributed.client.heartbeat': '5s',
             'distributed.client.preload': [], 'distributed.client.preload-argv': [],
             'distributed.client.scheduler-info-interval': '2s', 'distributed.client.security-loader': None,
             'distributed.comm.compression': False, 'distributed.comm.default-scheme': 'tcp',
             'distributed.comm.offload': '10MiB', 'distributed.comm.require-encryption': None,
             'distributed.comm.retry.count': 0, 'distributed.comm.retry.delay.max': '20s',
             'distributed.comm.retry.delay.min': '1s', 'distributed.comm.shard': '64MiB',
             'distributed.comm.socket-backlog': 2048, 'distributed.comm.timeouts.connect': '30s',
             'distributed.comm.timeouts.tcp': '30s', 'distributed.comm.tls.ca-file': None,
             'distributed.comm.tls.ciphers': None, 'distributed.comm.tls.client.cert': None,
             'distributed.comm.tls.client.key': None, 'distributed.comm.tls.max-version': None,
             'distributed.comm.tls.min-version': 1.2, 'distributed.comm.tls.scheduler.cert': None,
             'distributed.comm.tls.scheduler.key': None, 'distributed.comm.tls.worker.cert': None,
             'distributed.comm.tls.worker.key': None, 'distributed.comm.ucx.create-cuda-context': None,
             'distributed.comm.ucx.cuda-copy': None, 'distributed.comm.ucx.environment': {},
             'distributed.comm.ucx.infiniband': None, 'distributed.comm.ucx.nvlink': None,
             'distributed.comm.ucx.rdmacm': None, 'distributed.comm.ucx.tcp': None,
             'distributed.comm.websockets.shard': '8MiB', 'distributed.comm.zstd.level': 3,
             'distributed.comm.zstd.threads': 0, 'distributed.dashboard.export-tool': False,
             'distributed.dashboard.graph-max-items': 5000,
             'distributed.dashboard.link': '{scheme}://{host}:{port}/status',
             'distributed.dashboard.prometheus.namespace': 'dask', 'distributed.deploy.cluster-repr-interval': '500ms',
             'distributed.deploy.lost-worker-timeout': '15s',
             'distributed.diagnostics.computations.ignore-files': [r'runpy\.py', 'pytest', r'py\.test',
                                                                   r'pytest-script\.py', '_pytest', 'pycharm',
                                                                   'vscode_pytest', r'get_output_via_markers\.py'],
             'distributed.diagnostics.computations.ignore-modules': ['asyncio', 'functools', 'threading', 'datashader',
                                                                     'dask', 'debugpy', 'distributed', 'coiled', 'cudf',
                                                                     'cuml', 'matplotlib', 'pluggy', 'prefect',
                                                                     'rechunker', 'xarray', 'xgboost', 'xdist',
                                                                     '__channelexec__', 'execnet'],
             'distributed.diagnostics.computations.max-history': 100, 'distributed.diagnostics.computations.nframes': 0,
             'distributed.diagnostics.cudf': False, 'distributed.diagnostics.erred-tasks.max-history': 100,
             'distributed.diagnostics.nvml': True, 'distributed.nanny.environ': {},
             'distributed.nanny.pre-spawn-environ.MALLOC_TRIM_THRESHOLD_': 65536,
             'distributed.nanny.pre-spawn-environ.MKL_NUM_THREADS': 1,
             'distributed.nanny.pre-spawn-environ.OMP_NUM_THREADS': 1,
             'distributed.nanny.pre-spawn-environ.OPENBLAS_NUM_THREADS': 1, 'distributed.nanny.preload': [],
             'distributed.nanny.preload-argv': [], 'distributed.p2p.comm.buffer': '1 GiB',
             'distributed.p2p.comm.concurrency': 10, 'distributed.p2p.comm.message-bytes-limit': '2 MiB',
             'distributed.p2p.comm.retry.count': 10, 'distributed.p2p.comm.retry.delay.max': '30s',
             'distributed.p2p.comm.retry.delay.min': '1s', 'distributed.p2p.storage.buffer': '100 MiB',
             'distributed.p2p.storage.disk': True, 'distributed.p2p.threads': None, 'distributed.rmm.pool-size': None,
             'distributed.scheduler.active-memory-manager.interval': '2s',
             'distributed.scheduler.active-memory-manager.measure': 'optimistic',
             'distributed.scheduler.active-memory-manager.policies': [
                 {'class': 'distributed.active_memory_manager.ReduceReplicas'}],
             'distributed.scheduler.active-memory-manager.start': True, 'distributed.scheduler.allowed-failures': 3,
             'distributed.scheduler.allowed-imports': ['dask', 'distributed'],
             'distributed.scheduler.bandwidth': '100000000', 'distributed.scheduler.blocked-handlers': [],
             'distributed.scheduler.contact-address': None,
             'distributed.scheduler.dashboard.bokeh-application.allow_websocket_origin': ['*'],
             'distributed.scheduler.dashboard.bokeh-application.check_unused_sessions_milliseconds': 500,
             'distributed.scheduler.dashboard.bokeh-application.keep_alive_milliseconds': 500,
             'distributed.scheduler.dashboard.status.task-stream-length': 1000,
             'distributed.scheduler.dashboard.tasks.task-stream-length': 100000,
             'distributed.scheduler.dashboard.tls.ca-file': None, 'distributed.scheduler.dashboard.tls.cert': None,
             'distributed.scheduler.dashboard.tls.key': None, 'distributed.scheduler.default-data-size': '1kiB',
             'distributed.scheduler.default-task-durations.rechunk-split': '1us',
             'distributed.scheduler.default-task-durations.split-shuffle': '1us',
             'distributed.scheduler.default-task-durations.split-stage': '1us',
             'distributed.scheduler.default-task-durations.split-taskshuffle': '1us',
             'distributed.scheduler.events-cleanup-delay': '1h',
             'distributed.scheduler.http.routes': ['distributed.http.scheduler.prometheus',
                                                   'distributed.http.scheduler.info', 'distributed.http.scheduler.json',
                                                   'distributed.http.health', 'distributed.http.proxy',
                                                   'distributed.http.statics'],
             'distributed.scheduler.idle-timeout': None, 'distributed.scheduler.locks.lease-timeout': '30s',
             'distributed.scheduler.locks.lease-validation-interval': '10s',
             'distributed.scheduler.no-workers-timeout': None, 'distributed.scheduler.preload': [],
             'distributed.scheduler.preload-argv': [], 'distributed.scheduler.rootish-taskgroup': 5,
             'distributed.scheduler.rootish-taskgroup-dependencies': 5,
             'distributed.scheduler.unknown-task-duration': '500ms', 'distributed.scheduler.validate': False,
             'distributed.scheduler.work-stealing': True, 'distributed.scheduler.work-stealing-interval': '1s',
             'distributed.scheduler.worker-saturation': 1.1, 'distributed.scheduler.worker-ttl': '5 minutes',
             'distributed.version': 2, 'distributed.worker.blocked-handlers': [],
             'distributed.worker.connections.incoming': 10, 'distributed.worker.connections.outgoing': 50,
             'distributed.worker.daemon': True,
             'distributed.worker.http.routes': ['distributed.http.worker.prometheus', 'distributed.http.health',
                                                'distributed.http.statics'],
             'distributed.worker.lifetime.duration': None, 'distributed.worker.lifetime.restart': False,
             'distributed.worker.lifetime.stagger': '0 seconds', 'distributed.worker.memory.max-spill': False,
             'distributed.worker.memory.monitor-interval': '100ms', 'distributed.worker.memory.pause': 0.8,
             'distributed.worker.memory.rebalance.measure': 'optimistic',
             'distributed.worker.memory.rebalance.recipient-max': 0.6,
             'distributed.worker.memory.rebalance.sender-min': 0.3,
             'distributed.worker.memory.rebalance.sender-recipient-gap': 0.1,
             'distributed.worker.memory.recent-to-old-time': '30s', 'distributed.worker.memory.spill': 0.7,
             'distributed.worker.memory.spill-compression': 'auto', 'distributed.worker.memory.target': 0.6,
             'distributed.worker.memory.terminate': 0.95, 'distributed.worker.memory.transfer': 0.1,
             'distributed.worker.multiprocessing-method': 'spawn', 'distributed.worker.preload': [],
             'distributed.worker.preload-argv': [], 'distributed.worker.profile.cycle': '1000ms',
             'distributed.worker.profile.enabled': True, 'distributed.worker.profile.interval': '10ms',
             'distributed.worker.profile.low-level': False, 'distributed.worker.resources': {},
             'distributed.worker.transfer.message-bytes-limit': '50MB', 'distributed.worker.use-file-locking': True,
             'distributed.worker.validate': False
        }

        defaults = dict(
            cluster = dict(
                n_jobs=4,
                threads_per_worker=1,
                memory_limit='auto',
                temp_dir='auto',
                no_worker_restart=False,
                verbose=False,
                no_distributed=False,
                on_slurm = False,
            ),
            readers = dict(
                as_mosaic=False,
                view_index=0,
                phase_index=0,
                illumination_index=0,
                scene_index=0,
                rotation_index=0,
                mosaic_tile_index=0,
                sample_index=0
            ),                
            conversion = dict(
                zarr_format = 2,
                time_chunk = 1,
                channel_chunk = 1,
                z_chunk = 96,
                y_chunk=96,
                x_chunk=96,
                time_shard_coef = 1,
                channel_shard_coef = 1,
                z_shard_coef = 3,
                y_shard_coef = 3,
                x_shard_coef = 3,
                time_range = None,
                channel_range = None,
                z_range = None,
                y_range = None,
                x_range = None,
                dimension_order='tczyx',
                compressor = 'blosc',
                compressor_params = {},
                overwrite = False,
                use_tensorstore=False,
                use_gpu=False,
                rechunk_method='tasks',
                trim_memory=False,
                metadata_reader = 'bfio',
                save_omexml = True,
                squeeze = False
            ),
            downscale = dict(
                time_scale_factor = 1,
                channel_scale_factor = 1,
                z_scale_factor = 2,
                y_scale_factor = 2,
                x_scale_factor = 2,
                n_layers=3,
                downscale_method='simple',
            )
        )

        self.root_defaults = defaults
        self.root_dask_defaults = root_dask_defaults
        config_gr = zarr.open_group(configpath, mode = 'a')
        config = config_gr.attrs
        for key in defaults.keys():
            if key not in config.keys():
                config[key] = {}
                for subkey in defaults[key].keys():
                    if subkey not in config[key].keys():
                        config[key][subkey] = defaults[key][subkey]
            config_gr.attrs[key] = config[key]
        self.config = dict(config_gr.attrs)
        ###
        if not 'dask_config' in config_gr.keys():
            config_gr.create_group('dask_config')
        dask_config = config_gr['dask_config'].attrs
        for key in root_dask_defaults.keys():
            if key not in dask_config.keys():
                dask_config[key] = root_dask_defaults[key]
        config_gr['dask_config'].attrs.update(dict(dask_config))
        self.dask_config = dict(config_gr['dask_config'].attrs)
        self.config_gr = config_gr
        ###
        self._dask_temp_dir = None
        self.client = None

    def reset_config(self):
        """
        Resets the cluster, conversion and downscale parameters to the installation defaults.
        """
        self.config_gr.attrs.update(self.root_defaults)
        self.config = dict(self.config_gr.attrs)

    def reset_dask_config(self):
        """
        Resets the dask configuration parameters to the installation defaults.
        """
        self.config_gr['dask_config'].attrs.update(self.root_dask_defaults)
        self.dask_config = dict(self.config_gr['dask_config'].attrs)

    def show_config(self):
        """
        Displays the current cluster, conversion, and downscale parameters.
        """
        pprint.pprint(self.config)

    def show_dask_config(self):
        """
        Displays the current dask.distributed parameters.
        """
        pprint.pprint(self.dask_config)

    def show_root_defaults(self):
        """
        Displays the installation defaults for cluster, conversion, and downscale parameters.
        """
        pprint.pprint(self.root_defaults)

    def show_root_dask_defaults(self):
        """
        Displays the installation defaults for dask.distributed.
        """
        pprint.pprint(self.root_dask_defaults)

    def _collect_params(self, param_type, **kwargs):
        """
        Gathers parameters from the configuration, allowing for overrides.

        Args:
            param_type (str): The type of parameters to collect (e.g., 'cluster', 'conversion', 'downscale').
            **kwargs: Parameter values that may override defaults.

        Returns:
            dict: Collected parameters.
        """
        params = {}
        for key in self.config[param_type].keys():
            if key in kwargs.keys():
                params[key] = kwargs[key]
            else:
                params[key] = self.config[param_type][key]
        return params

    def configure_cluster(self,
                          memory_limit: str = 'default',
                          n_jobs: int = 'default',
                          no_worker_restart: bool = 'default',
                          on_slurm: bool = 'default',
                          temp_dir: str = 'default',
                          threads_per_worker: int = 'default',
                          no_distributed: bool = 'default',
                          verbose: bool = 'default'
                         ):
        """
        Updates cluster configuration settings. To update the current default value for a parameter, provide that parameter with a value other than 'default'.

        The following parameters can be configured:
            - memory_limit (str, optional): Memory limit per worker.
            - n_jobs (int, optional): Number of parallel jobs.
            - no_worker_restart (bool, optional): Whether to prevent worker restarts.
            - on_slurm (bool, optional): Whether running on a SLURM cluster.
            - temp_dir (str, optional): Temporary directory for Dask workers.
            - threads_per_worker (int, optional): Number of threads per worker.
            - verbose (bool, optional): Enables detailed logging.

        Args:
            memory_limit (str, optional): Memory limit per worker.
            n_jobs (int, optional): Number of parallel jobs.
            no_worker_restart (bool, optional): Whether to prevent worker restarts.
            on_slurm (bool, optional): Whether running on a SLURM cluster.
            temp_dir (str, optional): Temporary directory for Dask workers.
            threads_per_worker (int, optional): Number of threads per worker.
            verbose (bool, optional): Enables detailed logging.

        Returns:
            None
        """

        params = {
            'memory_limit': memory_limit,
            'n_jobs': n_jobs,
            'no_worker_restart': no_worker_restart,
            'on_slurm': on_slurm,
            'temp_dir': temp_dir,
            'threads_per_worker': threads_per_worker,
            'no_distributed': no_distributed,
            'verbose': verbose
        }

        for key in params:
            if key in self.config['cluster'].keys():
                if params[key] != 'default':
                    self.config['cluster'][key] = params[key]
        self.config_gr.attrs['cluster'] = self.config['cluster']

    def configure_readers(self,
                          as_mosaic: bool = 'default',
                          view_index: int = 'default',
                          phase_index: int = 'default',
                          illumination_index: int = 'default',
                          scene_index: int = 'default',
                          rotation_index: int = 'default',
                          mosaic_tile_index: int = 'default',
                          sample_index: int = 'default',
                         ):
        """
        Updates reader configuration settings. To update the current default value for a parameter, provide that parameter with a value other than 'default'.

        Returns:
            None
        """

        params = {
            'as_mosaic': as_mosaic,
            'view_index': view_index,
            'phase_index': phase_index,
            'illumination_index': illumination_index,
            'scene_index': scene_index,
            'rotation_index': rotation_index,
            'mosaic_tile_index': mosaic_tile_index,
            'sample_index': sample_index
        }

        for key in params:
            if key in self.config['readers'].keys():
                if params[key] != 'default':
                    self.config['readers'][key] = params[key]
        self.config_gr.attrs['readers'] = self.config['readers']


    def configure_conversion(self,
                             zarr_format: int = 'default',
                             compressor: str = 'default',
                             compressor_params: dict = 'default',
                             time_chunk: int = 'default',
                             channel_chunk: int = 'default',
                             z_chunk: int = 'default',
                             y_chunk: int = 'default',
                             x_chunk: int = 'default',
                             time_shard_coef: int = 'default',
                             channel_shard_coef: int = 'default',
                             z_shard_coef: int = 'default',
                             y_shard_coef: int = 'default',
                             x_shard_coef: int = 'default',
                             time_range: int = 'default',
                             channel_range: int = 'default',
                             z_range: int = 'default',
                             y_range: int = 'default',
                             x_range: int = 'default',
                             dimension_order: str = 'default',
                             overwrite: bool = 'default',
                             rechunk_method: str = 'default',
                             # rechunkers_max_mem: str = 'default',
                             trim_memory: bool = 'default',
                             use_tensorstore: bool = 'default',
                             use_gpu: bool = 'default',
                             metadata_reader: str = 'default',
                             save_omexml: bool = 'default',
                             squeeze: bool = 'default'
                             ):
        """
        Updates conversion configuration settings. To update the current default value for a parameter, provide that parameter with a value other than 'default'.

        The following parameters can be configured:
            - compressor (str, optional): Compression algorithm.
            - compressor_params (dict, optional): Parameters for the compressor.
            - output_chunks (list, optional): Chunk size for output.
            - overwrite (bool, optional): Whether to overwrite existing data.
            - rechunk_method (str, optional): Method used for rechunking.
            - trim_memory (bool, optional): Whether to trim memory usage.
            - use_tensorstore (bool, optional): Whether to use TensorStore for writing.
            - save_omexml (bool, optional): Whether to create a METADATA.ome.xml file.
        Args:
            compressor (str, optional): Compression algorithm.
            compressor_params (dict, optional): Parameters for the compressor.
            output_chunks (list, optional): Chunk size for output.
            overwrite (bool, optional): Whether to overwrite existing data.
            rechunk_method (str, optional): Method used for rechunking.
            trim_memory (bool, optional): Whether to trim memory usage.
            use_tensorstore (bool, optional): Whether to use TensorStore for storage.
            save_omexml (bool, optional): Whether to create a METADATA.ome.xml file.

        Returns:
            None
        """

        params = {
            'zarr_format': zarr_format,
            'compressor': compressor,
            'compressor_params': compressor_params,
            "time_chunk": time_chunk,
            "channel_chunk": channel_chunk,
            "z_chunk": z_chunk,
            "y_chunk": y_chunk,
            "x_chunk": x_chunk,
            "time_shard_coef": time_shard_coef,
            "channel_shard_coef": channel_shard_coef,
            "z_shard_coef": z_shard_coef,
            "y_shard_coef": y_shard_coef,
            "x_shard_coef": x_shard_coef,
            "time_range": time_range,
            "channel_range": channel_range,
            "z_range": z_range,
            "y_range": y_range,
            "x_range": x_range,
            "dimension_order": dimension_order,
            'overwrite': overwrite,
            'rechunk_method': rechunk_method,
            # 'rechunkers_max_mem': rechunkers_max_mem,
            'trim_memory': trim_memory,
            'use_tensorstore': use_tensorstore,
            'use_gpu': use_gpu,
            'metadata_reader': metadata_reader,
            'save_omexml': save_omexml,
            'squeeze': squeeze
        }

        for key in params:
            if key in self.config['conversion'].keys():
                if params[key] != 'default':
                    self.config['conversion'][key] = params[key]
        self.config_gr.attrs['conversion'] = self.config['conversion']

    def configure_downscale(self,
                            downscale_method: str = 'default',
                            n_layers: int = 'default',
                            time_scale_factor: int = 'default',
                            channel_scale_factor: int = 'default',
                            z_scale_factor: int = 'default',
                            y_scale_factor: int = 'default',
                            x_scale_factor: int = 'default',
                            ):
        """
        Updates downscaling configuration settings. To update the current default value for a parameter, provide that parameter with a value other than 'default'.

        The following parameters can be configured:
            - downscale_method (str, optional): Downscaling algorithm.
            - n_layers (int, optional): Number of downscaling layers.
            - scale_factor (list, optional): Scaling factors for each dimension.

        Args:
            downscale_method (str, optional): Downscaling algorithm.
            n_layers (int, optional): Number of downscaling layers.
            scale_factor (list, optional): Scaling factors for each dimension.

        Returns:
            None
        """

        params = {
            'downscale_method': downscale_method,
            'n_layers': n_layers,
            'time_scale_factor': time_scale_factor,
            "channel_scale_factor": channel_scale_factor,
            "z_scale_factor": z_scale_factor,
            "y_scale_factor": y_scale_factor,
            "x_scale_factor": x_scale_factor,
        }

        for key in params:
            if key in self.config['downscale'].keys():
                if params[key] != 'default':
                    self.config['downscale'][key] = params[key]
        self.config_gr.attrs['downscale'] = self.config['downscale']

    def _set_dask_temp_dir(self, temp_dir = 'auto'):
        if self._dask_temp_dir is not None:
            self._dask_temp_dir.cleanup()
        if temp_dir in ('auto', None):
            temp_dir = tempfile.TemporaryDirectory()
        else:
            os.makedirs(temp_dir, exist_ok=True)
            temp_dir = tempfile.TemporaryDirectory(dir=temp_dir)
        self._dask_temp_dir = temp_dir
        return self

    def _start_cluster(self,
                    n_jobs: int = 4,
                    threads_per_worker: int = 1,
                    memory_limit: str = 'auto',
                    temp_dir='auto',
                    verbose = False,
                    on_slurm = False,
                    no_distributed = False,
                    config_kwargs = {},
                    **kwargs
                    ):

        config_dict = copy.deepcopy(self.dask_config)
        config_dict.update(**config_kwargs)

        # worker_options = {
        #     "memory_target_fraction": 0.8,
        #     "memory_spill_fraction": 0.9,
        #     "memory_pause_fraction": 0.95,
        #     # "memory_terminate_fraction": 0.98
        # }
        scheduler_options = {
            "allowed_failures": 100,
            "idle_timeout": "1h",
            "worker_ttl": "1d"  # Set to a large value, e.g., 1 day
        }

        self._set_dask_temp_dir(temp_dir)

        dask.config.set(config_dict) # use dictionary notation here.

        if no_distributed:
            config_dict.update(scheduler = 'threads',
                               pool = ThreadPool(n_jobs)
                               )
            dask.config.set(config_dict)
            logger.info(f"Process running locally via multithreading.")
        else:
            if memory_limit == 'auto':
                reserve_fraction = kwargs.get('reserve_memory_fraction', 0.1)
                min_per_worker = kwargs.get('min_memory_per_worker', 1 * 1024 ** 3)

                total_mem = psutil.virtual_memory().total
                reserved_mem = total_mem * reserve_fraction
                available_mem = max(0, total_mem - reserved_mem)
                mem_per_worker = max(available_mem / n_jobs, min_per_worker)
                mem_gb = mem_per_worker / (1 * 1024 ** 3)
                memory_limit = f"{mem_gb} GB"
                logger.info(f"{memory_limit} memory has been allocated per worker.")

            if on_slurm:
                logger.info(f"Process running on Slurm.")
                cluster = SLURMCluster(
                                        cores=threads_per_worker,
                                        processes=1,
                                        nanny=False,
                                        scheduler_options=scheduler_options,
                                        n_workers=n_jobs,
                                        memory=memory_limit,
                                        local_directory=f"{self._dask_temp_dir.name}",
                                        # **worker_options
                                        )
            else:
                logger.info(f"Process running on local cluster.")
                cluster = LocalCluster(
                                       n_workers=n_jobs,
                                       threads_per_worker=threads_per_worker,
                                       nanny = False,
                                       scheduler_kwargs=scheduler_options,
                                       memory_limit = memory_limit,
                                       local_directory=f"{self._dask_temp_dir.name}",
                                       # **worker_options
                                       )
            cluster.scale(n_jobs)
            self.client = Client(cluster)
            if verbose:
                logger.info(self.client.cluster)
        return self

    def to_zarr(self,
                input_path: Union[Path, str],
                output_path: Union[Path, str],
                includes=None,
                excludes=None,
                time_tag: Union[str, tuple] = None,
                channel_tag: Union[str, tuple] = None,
                z_tag: Union[str, tuple] = None,
                y_tag: Union[str, tuple] = None,
                x_tag: Union[str, tuple] = None,
                concatenation_axes: Union[int, tuple, str] = None,
                **kwargs
                ):
        """
        Converts image data to OME-Zarr format and optionally applies downscaling.

        Args:
            input_path (Union[Path, str]): Path to input file or directory.
            output_path (Union[Path, str]): Directory, in which the output OME-Zarrs will be written.
            includes (str, optional): Filename patterns to filter for.
            excludes (str, optional): Filename patterns to filter against.
            time_tag (Union[str, tuple], optional): Time dimension tag.
            channel_tag (Union[str, tuple], optional): Channel dimension tag.
            z_tag (Union[str, tuple], optional): Z dimension tag.
            y_tag (Union[str, tuple], optional): Y dimension tag.
            x_tag (Union[str, tuple], optional): X dimension tag.
            concatenation_axes (Union[int, tuple, str], optional): Axes, along which the images will be concatenated.
            **kwargs: Additional configuration overrides.

        Raises:
            Exception: If no files are found in the input path.

        Prints:
            Process logs including conversion and downscaling time.

        Returns:
            None
        """
        t0 = time.time()
        # Get parameters:
        self.cluster_params = self._collect_params('cluster', **kwargs)
        self.readers_params = self._collect_params('readers', **kwargs)
        self.conversion_params = self._collect_params('conversion', **kwargs)
        self.downscale_params = self._collect_params('downscale', **kwargs)

        if self.conversion_params['use_gpu'] and self.conversion_params['use_tensorstore']:
            raise ValueError("Tensorstore is not supported for GPU arrays.")

        logger.info(f"Base conversion initiated.")
        ###### Handle input data and metadata
        paths = take_filepaths(input_path, includes = includes, excludes = excludes)

        filepaths = sorted(list(paths))

        ###### Start the cluster
        verified_for_cluster = verify_filepaths_for_cluster(filepaths) ### Ensure non-bioformats conversion. If bioformats is needed, fall back on local conversion.
        if not verified_for_cluster:
            self.cluster_params['no_distributed'] = True

        cluster_is_true = not self.cluster_params['no_distributed']

        if cluster_is_true:
            chunks_yx = None
        else:
            chunks_yx = tuple([self.conversion_params['y_chunk'],  # chunks already when reading
                               self.conversion_params['x_chunk']]
                              )

        self._start_cluster(**self.cluster_params)


        series = self.readers_params['scene_index']

        ###### Read and concatenate
        base = BridgeBase(input_path,
                        excludes=excludes,
                        includes=includes,
                        series=series,
                        zarr_format = self.conversion_params['zarr_format'],
                        verbose = self.cluster_params['verbose']
                        )

        base.read_dataset(cluster_is_true,
                          chunks_yx=chunks_yx,
                          readers_params = self.readers_params
                          )

        base.digest(
            time_tag = time_tag,
            channel_tag = channel_tag,
            z_tag = z_tag,
            y_tag = y_tag,
            x_tag = x_tag,
            axes_of_concatenation = concatenation_axes,
            metadata_reader = self.conversion_params['metadata_reader'],
            **kwargs
            )
        logger.info(f"Metadata was extracted")
        verbose = base._verbose

        if 'region_shape' in kwargs.keys():
            self.conversion_params['region_shape'] = kwargs.get('region_shape')
        if verbose:
            print(f"Cluster params:")
            pprint.pprint(self.cluster_params)
            print(f"Readers params:")
            pprint.pprint(self.readers_params)
            print(f"Conversion params:")
            pprint.pprint(self.conversion_params)
            print(f"Downscale params:")
            pprint.pprint(self.downscale_params)

        temp_dir = base._dask_temp_dir
        self.conversion_params['temp_dir'] = temp_dir
        self.downscale_params['temp_dir'] = temp_dir

        if self.client is not None:
            base.client = self.client
        base.set_dask_temp_dir(self._dask_temp_dir)

        ###### Write
        self.base_results = base.write_arrays(output_path,
                                           compute=True,
                                           verbose=verbose,
                                           **self.conversion_params
                                           )
        ###### Downscale
        logger.info(f"Base conversion finished.")
        t1 = time.time()
        logger.info(f"Elapsed for base conversion: {(t1 - t0) / 60} min.")
        n_layers = self.downscale_params['n_layers']
        if n_layers > 1:
            logger.info(f"Downscaling initiated.")
            _ = downscale(
                      self.base_results,
                      **self.downscale_params,
                      zarr_format = self.conversion_params['zarr_format'],
                      rechunk_method = self.conversion_params['rechunk_method'],
                      use_tensorstore = self.conversion_params['use_tensorstore'],
                      verbose = verbose
                      ) # TODO: add to_cupy parameter here.

            logger.info(f"Downscaling finished.")

        ###### Shutdown and clean up
        if self.client is not None:
            self.client.shutdown()
            self.client.close()

        if isinstance(self._dask_temp_dir, tempfile.TemporaryDirectory):
            shutil.rmtree(self._dask_temp_dir.name)
        else:
            shutil.rmtree(self._dask_temp_dir)

        t1 = time.time()
        logger.info(f"Elapsed for conversion + downscaling: {(t1 - t0) / 60} min.")


    def show_pixel_meta(self,
                input_path: Union[Path, str],
                includes=None,
                excludes=None,
                series: int = None, # self.readers_params['scene_index'],
                **kwargs
                ):
        """
        Print pixel-level metadata for all datasets in the 'input_path'.

        Args:
            input_path (Union[Path, str]): Path to input file or directory.
            output_path (Union[Path, str]): Directory, in which the output OME-Zarrs will be written.
            includes (str, optional): Filename patterns to filter for.
            excludes (str, optional): Filename patterns to filter against.
            **kwargs: Additional configuration overrides.

        Raises:
            Exception: If no files are found in the input path.

        Prints:
            Process logs including conversion and downscaling time.

        Returns:
            None
        """

        # Get parameters:
        self.cluster_params = self._collect_params('cluster', **kwargs)
        self.readers_params = self._collect_params('readers', **kwargs)
        self.conversion_params = self._collect_params('conversion', **kwargs)

        paths = take_filepaths(input_path, includes = includes, excludes = excludes)

        filepaths = sorted(list(paths))

        ###### Start the cluster
        verified_for_cluster = verify_filepaths_for_cluster(filepaths)
        if not verified_for_cluster:
            self.cluster_params['no_distributed'] = True

        self._start_cluster(**self.cluster_params)

        series = self.readers_params['scene_index']

        ###### Read and digest
        base = BridgeBase(input_path,
                        excludes=excludes,
                        includes=includes,
                        series=series
                        )

        base.read_dataset(verified_for_cluster,
                          readers_params=self.readers_params
                          )

        base.digest()

        temp_dir = base._dask_temp_dir
        self.conversion_params['temp_dir'] = temp_dir

        if self.client is not None:
            base.client = self.client
        base.set_dask_temp_dir(self._dask_temp_dir)

        printables = {
            path: get_printables(
                manager.axes,
                manager.shapedict,
                manager.scaledict,
                manager.unitdict
            )
            for path, manager in base.batchdata.managers.items()
        }
        for path, printable in printables.items():
            print('---------')
            print(f"")
            print(f"Metadata for '{path}':")
            print_printable(printable)

        ###### Shutdown and clean up
        if self.client is not None:
            self.client.shutdown()
            self.client.close()

        if isinstance(self._dask_temp_dir, tempfile.TemporaryDirectory):
            shutil.rmtree(self._dask_temp_dir.name)
        else:
            shutil.rmtree(self._dask_temp_dir)

    def update_pixel_meta(self,
                          input_path: Union[Path, str],
                          includes=None,
                          excludes=None,
                          time_scale: (int, float) = None,
                          z_scale: (int, float) = None,
                          y_scale: (int, float) = None,
                          x_scale: (int, float) = None,
                          time_unit: str = None,
                          z_unit: str = None,
                          y_unit: str = None,
                          x_unit: str = None,
                          **kwargs
                          ):
        """
        Updates pixel metadata for image files located at the specified input path.

        Args:
            input_path (Union[Path, str]): Path to input file or directory.
            includes (optional): Filename patterns to include.
            excludes (optional): Filename patterns to exclude.
            series (int, optional): Series index to process.
            time_scale, z_scale, y_scale, x_scale ((int, float), optional): Scaling factors for the respective dimensions.
            time_unit, z_unit, y_unit, x_unit (str, optional): Units for the respective dimensions.
            **kwargs: Additional parameters for cluster and conversion configuration.

        Returns:
            None
        """

        # Collect cluster and conversion parameters
        self.cluster_params = self._collect_params('cluster', **kwargs)
        self.readers_params = self._collect_params('readers', **kwargs)
        self.conversion_params = self._collect_params('conversion', **kwargs)

        # Collect file paths based on inclusion and exclusion patterns
        paths = take_filepaths(input_path, includes=includes, excludes=excludes)

        filepaths = sorted(list(paths))

        # Verify file paths for cluster compatibility
        verified_for_cluster = verify_filepaths_for_cluster(filepaths)
        if not verified_for_cluster:
            self.cluster_params['no_distributed'] = True

        cluster_is_true = not self.cluster_params['no_distributed']

        # Start the processing cluster
        self._start_cluster(**self.cluster_params)

        series = self.readers_params['scene_index']

        base = BridgeBase(input_path,
                          excludes=excludes,
                          includes=includes,
                          series=series
                          )

        # Read and digest the dataset
        base.read_dataset(verified_for_cluster,
                          readers_params=self.readers_params
                          )

        # Prepare pixel metadata arguments
        pixel_meta_kwargs_ = dict(time_scale=time_scale,
                                  z_scale=z_scale,
                                  y_scale=y_scale,
                                  x_scale=x_scale,
                                  time_unit=time_unit,
                                  z_unit=z_unit,
                                  y_unit=y_unit,
                                  x_unit=x_unit)
        pixel_meta_kwargs = {key: val for key, val in pixel_meta_kwargs_.items() if val is not None}

        base.digest(**pixel_meta_kwargs)
        logger.info(f"Metadata was extracted")

        if self.client is not None:
            base.client = self.client
        base.set_dask_temp_dir(self._dask_temp_dir)

        # Update metadata for each dataset manager
        for path, manager in base.batchdata.managers.items():
            if is_zarr_group(manager.path):
                manager.sync_pyramid(self.conversion_params['save_omexml'])
            else:
                logger.info(f"Cannot update metadata for non-zarr path: {path}")

        # Shutdown the cluster and clean up temporary directories
        if self.client is not None:
            self.client.shutdown()
            self.client.close()

        if isinstance(self._dask_temp_dir, tempfile.TemporaryDirectory):
            shutil.rmtree(self._dask_temp_dir.name)
        else:
            shutil.rmtree(self._dask_temp_dir)



